"""HSSM - Hierarchical Sequential Sampling Models."""

import logging
import sys

from .config import show_defaults
from .datasets import load_data
from .hssm import HSSM
from .simulator import simulate_data
from .utils import set_floatX

_logger = logging.getLogger("hssm")
_logger.setLevel(logging.INFO)
handler = logging.StreamHandler(stream=sys.stdout)
_logger.addHandler(handler)

__all__ = ["HSSM", "load_data", "simulate_data", "set_floatX", "show_defaults"]



"""The Param utility class."""

from __future__ import annotations

from typing import Any, Callable, Union, cast

import bambi as bmb
import numpy as np
import pymc as pm
from bambi.backend.utils import get_distribution

# PEP604 union operator "|" not supported by pylint
# Fall back to old syntax

# Explicitly define types so they are more expressive
# and reusable
ParamSpec = Union[float, dict[str, Any], bmb.Prior]
BoundsSpec = tuple[float, float]


class Param:
    """Represents the specifications for the main HSSM class.

    Also provides convenience functions that can be used by the HSSM class to parse
    arguments.

    Parameters
    ----------
    name
        The name of the parameter.
    prior
        If a formula is not specified (the non-regression case), this parameter
        expects a float value if the parameter is fixed or a dictionary that can be
        parsed by Bambi as a prior specification or a Bambi Prior object. If not
        specified, then a default uninformative uniform prior with `bound` as
        boundaries will be constructed. An error will be thrown if `bound` is also
        not specified.
        If a formula is specified (the regression case), this parameter expects a
        dictionary of param:prior, where param is the name of the response variable
        specified in formula, and prior is specified as above. If left unspecified,
        default priors created by Bambi will be used.
    formula
        The regression formula if the parameter depends on other variables. The
        response variable can be omitted.
    link
        The link function for the regression. It is either a string that specifies
        a built-in link function in Bambi, or a Bambi Link object. If a regression
        is specified and link is not specified, "identity" will be used by default.
    bounds
        If provided, the prior will be created with boundary checks. If this
        parameter is specified as a regression, boundary checks will be skipped at
        this point.
    is_parent
        Determines if this parameter is a "parent" parameter. If so, the response
        term for the formula will be "c(rt, response)". Default is False.
    """

    def __init__(
        self,
        name: str,
        prior: ParamSpec | dict[str, ParamSpec] | None = None,
        formula: str | None = None,
        link: str | bmb.Link | None = None,
        bounds: BoundsSpec | None = None,
        is_parent: bool = False,
    ):
        self.name = name
        self.formula = formula
        self._parent = is_parent
        self.bounds = tuple(float(x) for x in bounds) if bounds is not None else None
        self._is_truncated = False

        if self.bounds is not None:
            self.bounds = cast(BoundsSpec, self.bounds)
            if any(not np.isscalar(bound) for bound in self.bounds):
                raise ValueError(f"The bounds of {self.name} should both be scalar.")
            lower, upper = self.bounds
            assert lower < upper, (
                f"The lower bound of {self.name} should be less than "
                + "its upper bound."
            )

        if isinstance(prior, int):
            prior = float(prior)

        if formula is not None:
            # The regression case

            self.formula = formula if "~" in formula else f"{name} ~ {formula}"

            if isinstance(prior, (float, bmb.Prior)):
                raise ValueError(
                    "Please specify priors for each individual parameter in the "
                    + "regression."
                )

            self.prior: float | bmb.Prior = (
                _make_prior_dict(prior) if prior is not None else prior
            )

            self.link = "identity" if link is None else link

        else:
            # The non-regression case

            if prior is None:
                if self.bounds is None:
                    raise ValueError(
                        f"Please specify the prior or bounds for {self.name}."
                    )
                self.prior = _make_default_prior(self.bounds)
            else:
                # Explicitly cast the type of prior, no runtime performance penalty
                prior = cast(ParamSpec, prior)

                if self.bounds is None:
                    if isinstance(prior, (float, bmb.Prior)):
                        self.prior = prior
                    else:
                        self.prior = bmb.Prior(**prior)
                else:
                    if isinstance(prior, float):
                        self.prior = prior
                    else:
                        self.prior = make_bounded_prior(prior, self.bounds)
                        # self._prior is internally used for informative output
                        # Not used in inference
                        self._prior = (
                            bmb.Prior(**prior) if isinstance(prior, dict) else prior
                        )
                        self._is_truncated = True

            if link is not None:
                raise ValueError("`link` should be None if no regression is specified.")

            self.link = None

    @property
    def is_regression(self) -> bool:
        """Determines if a regression is specified or not.

        Returns
        -------
        bool
            A boolean that indicates if a regression is specified.
        """
        return self.formula is not None

    @property
    def is_parent(self) -> bool:
        """Determines if a parameter is a parent parameter for Bambi.

        Returns
        -------
        bool
            A boolean that indicates if the parameter is a parent or not.
        """
        return self._parent

    @property
    def is_fixed(self) -> bool:
        """Determine if a parameter is a fixed value.

        Returns
        -------
        bool
            A boolean that indicates if the parameter is a fixed value.
        """
        return isinstance(self.prior, float)

    @property
    def is_truncated(self) -> bool:
        """Determines if a parameter is truncated.

        A parameter is truncated when it is not a regression, is not fixed and has
        bounds.

        Returns
        -------
            A boolean that indicates if a parameter is truncated.
        """
        return self._is_truncated

    def _parse_bambi(
        self,
    ) -> tuple:
        """
        Return a 3-tuple that helps with constructing the Bambi model.

        Returns
        -------
        tuple
            A 3-tuple of formula, priors, and link functions that can be used to
            construct the Bambi model.
        """
        formula = None
        prior = None
        link = None

        # Again, to satisfy type checker
        # Equivalent to `if self.is_regression`
        if self.formula is not None:
            left_side = "c(rt, response)" if self._parent else self.name

            right_side = self.formula.split("~")[1]
            right_side = right_side.strip()
            formula = f"{left_side} ~ {right_side}"

            if self.prior is not None:
                prior = {left_side: self.prior}
            link = {self.name: self.link}

            return formula, prior, link

        formula = "c(rt, response) ~ 1" if self._parent else None

        if self._parent:
            prior = {"c(rt, response)": {"Intercept": self.prior}}
            link = {self.name: "identity"}
        else:
            prior = {self.name: self.prior}  # type: ignore

        return formula, prior, link

    def __repr__(self) -> str:
        """Return the representation of the class.

        Returns
        -------
        str
            A string whose construction depends on whether the specification contains a
            regression or not.
        """
        output = []
        output.append(self.name + ":")

        # Simplest case: float
        # Output Value: 0.2
        if isinstance(self.prior, float):
            output.append(f"    Value: {self.prior}")
            return "\r\n".join(output)

        # Regression case:
        # Output formula, priors, and link functions
        if self.is_regression:
            assert self.formula is not None
            output.append(f"    Formula: {self.formula}")
            output.append("    Priors:")

            if self.prior is not None:
                assert isinstance(self.prior, dict)

                for param, prior in self.prior.items():
                    output.append(f"        {param} ~ {prior}")
            else:
                output.append("        Unspecified. Using defaults")

            link = self.link if isinstance(self.link, str) else self.link.name
            output.append(f"    Link: {link}")

        # None regression case:
        # Output prior and bounds
        else:
            assert isinstance(self.prior, bmb.Prior)
            prior_output = self._prior if self.is_truncated else self.prior
            output.append(f"    Prior: {prior_output}")

        output.append(f"    Explicit bounds: {self.bounds}")
        return "\r\n".join(output)

    def __str__(self) -> str:
        """Return the string representation of the class.

        Returns
        -------
        str
            A string whose construction depends on whether the specification contains a
            regression or not.
        """
        return self.__repr__()


def _make_prior_dict(prior: dict[str, ParamSpec]) -> dict[str, float | bmb.Prior]:
    """Make bambi priors from a ``dict`` of priors for the regression case.

    Parameters
    ----------
    prior
        A dictionary where each key is the name of a parameter in a regression
        and each value is the prior specification for that parameter.

    Returns
    -------
    dict[str, float | bmb.Prior]
        A dictionary where each key is the name of a parameter in a regression and each
        value is either a float or a bmb.Prior object.
    """
    priors = {
        # Convert dict to bmb.Prior if a dict is passed
        param: _make_priors_recursive(prior) if isinstance(prior, dict) else prior
        for param, prior in prior.items()
    }

    return priors


def _make_priors_recursive(prior: dict[str, Any]) -> bmb.Prior:
    """Make `bmb.Prior` objects from ``dict``s.

    Helper function that recursively converts a dict that might have some fields that
    have a parameter definitions as dicts to bmb.Prior objects.

    Parameters
    ----------
    prior
        A dictionary that contains parameter specifications.

    Returns
    -------
    bmb.Prior
        A bmb.Prior object with fields that can be converted to bmb.Prior objects also
        converted.
    """
    for k, v in prior.items():
        if isinstance(v, dict) and "name" in v:
            prior[k] = _make_priors_recursive(v)

    return bmb.Prior(**prior)


def _parse_bambi(
    params: list[Param],
) -> tuple[bmb.Formula, dict | None, dict[str, str | bmb.Link] | str]:
    """From a list of Params, retrieve three items that helps with bambi model building.

    Parameters
    ----------
    params
        A list of Param objects.

    Returns
    -------
    tuple
        A tuple containing:
            1. A bmb.Formula object.
            2. A dictionary of priors, if any is specified.
            3. A dictionary of link functions, if any is specified.
    """
    # Handle the edge case where list_params is empty:
    if not params:
        return bmb.Formula("c(rt, response) ~ 1"), None, "identity"

    # Then, we check how many parameters in the specified list of params are parent.
    num_parents = sum(param.is_parent for param in params)

    # In the case where there is more than one parent:
    assert num_parents <= 1, "More than one parent is specified!"

    formulas = []
    priors: dict[str, Any] = {}
    links: dict[str, str | bmb.Link] = {}
    params_copy = params.copy()

    parent_param = None

    if num_parents == 1:
        for idx, param in enumerate(params):
            if param.is_parent:
                parent_param = params_copy.pop(idx)
                break

        assert parent_param is not None
        params_copy.insert(0, parent_param)

    for param in params_copy:
        formula, prior, link = param._parse_bambi()

        if formula is not None:
            formulas.append(formula)
        if prior is not None:
            priors |= prior
        if link is not None:
            links |= link

    result_formula: bmb.Formula = (
        bmb.Formula("c(rt, response) ~ 1", *formulas)
        if num_parents == 0
        else bmb.Formula(formulas[0], *formulas[1:])
    )
    result_priors = None if not priors else priors

    result_links: dict | str = "identity" if not links else links

    return result_formula, result_priors, result_links


def make_bounded_prior(prior: ParamSpec, bounds: BoundsSpec) -> float | bmb.Prior:
    """Create prior within specific bounds.

    Helper function that creates a prior within specified bounds. Works in the
    following cases:

    1. If a prior passed is a fixed value, then check if the value is specified within
    the bounds. Raises a ValueError if not.
    2. If a prior passed is a dictionary, we create a bmb.Prior with a truncated
    distribution.
    3. If a prior is passed as a bmb.Prior object, do the same thing above.

    The above boundary checks do not happen when bounds is None.

    Parameters
    ----------
    prior
        A prior definition. Could be a float, a dict that can be passed to a bmb.Prior
        to create a prior distribution, or a bmb.Prior.
    bounds : optional
        If provided, needs to be a tuple of floats that indicates the lower and upper
        bounds of the parameter.

    Returns
    -------
    float | bmb.Prior
        A float if `prior` is a float, otherwise a bmb.Prior object.
    """
    lower, upper = bounds

    if isinstance(prior, float):
        if not lower <= prior <= upper:
            raise ValueError(
                f"The fixed prior should be between {lower:.4f} and {upper:.4f}, "
                + f"got {prior:.4f}."
            )

        return prior

    if isinstance(prior, dict):
        if np.isinf(lower) and np.isinf(upper):
            return bmb.Prior(**prior)
        dist = make_truncated_dist(lower, upper, **prior)
        return bmb.Prior(name=prior["name"], dist=dist)

    # After handling the constant and dict case, now handle the bmb.Prior case
    if np.isinf(lower) and np.isinf(upper):
        return prior
    if prior.dist is not None:
        return prior

    name = prior.name
    args = prior.args

    dist = make_truncated_dist(lower, upper, name=name, **args)

    return bmb.Prior(name=name, dist=dist)


def make_truncated_dist(lower_bound: float, upper_bound: float, **kwargs) -> Callable:
    """Create custom functions with truncated priors.

    Helper function that creates a custom function with truncated priors.

    Parameters
    ----------
    lower_bound
        The lower bound for the distribution.
    upper_bound
        The upper bound for the distribution.
    kwargs
        Typically a dictionary with a name for the name of the Prior distribution
        and other arguments passed to bmb.Prior object.

    Returns
    -------
    Callable
        A distribution (TensorVariable) created with pm.Truncated().
    """
    dist_name = kwargs["name"]
    dist_kwargs = {k: v for k, v in kwargs.items() if k != "name"}

    initval = dist_kwargs.pop("initval") if "initval" in dist_kwargs else None

    def TruncatedDist(name):
        dist = get_distribution(dist_name).dist(**dist_kwargs)
        return pm.Truncated(
            name=name,
            dist=dist,
            lower=lower_bound if np.isfinite(lower_bound) else None,
            upper=upper_bound if np.isfinite(upper_bound) else None,
            initval=initval,
        )

    return TruncatedDist


def _make_default_prior(bounds: tuple[float, float]) -> bmb.Prior:
    """Make a default prior from bounds.

    Parameters
    ----------
    bounds
        The (lower, upper) bounds for the default prior.

    Returns
    -------
        A bmb.Prior object representing the default prior for the provided bounds.
    """
    lower, upper = bounds
    if np.isinf(lower) and np.isinf(upper):
        return bmb.Prior("Normal", mu=0.0, sigma=2.0)
    elif np.isinf(lower) and not np.isinf(upper):
        return bmb.Prior("TruncatedNormal", mu=upper, upper=upper, sigma=2.0)
    elif not np.isinf(lower) and np.isinf(upper):
        if lower == 0:
            return bmb.Prior("HalfNormal", sigma=2.0)
        return bmb.Prior("TruncatedNormal", mu=lower, lower=lower, sigma=2.0)
    else:
        return bmb.Prior(name="Uniform", lower=lower, upper=upper)



"""
Base IO code for datasets.

Heavily influenced by Arviz's(scikit-learn's, and Bambi's) implementation.
"""

import os
from collections import namedtuple
from typing import Optional, Union

import pandas as pd

base_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

FileMetadata = namedtuple("FileMetadata", ["filename", "path", "description"])

DATASETS = {
    "cavanagh_theta": FileMetadata(
        filename="cavanagh_theta",
        path=os.path.join(base_dir, "hssm/datasets/cavanagh_theta_nn.csv"),
        description="Description for cavanagh_theta dataset",
    )
}


def load_data(dataset: Optional[str] = None) -> Union[pd.DataFrame, str]:
    """
    Load a dataset as a pandas DataFrame.

    If a valid dataset name is provided, this function will return the
    corresponding DataFrame. Otherwise, it lists the available datasets.

    Parameters
    ----------
    dataset : str, optional
        Name of the dataset to load. If not provided, a list
        of available datasets is returned.

    Raises
    ------
    ValueError
        If the provided dataset name does not match any of the available datasets.

    Returns
    -------
    pd.DataFrame or str
        Loaded dataset as a DataFrame if a valid dataset name was provided,
        otherwise a string listing the available datasets.
    """
    if dataset in DATASETS:
        datafile = DATASETS[dataset]
        file_path = datafile.path

        if not os.path.exists(file_path):
            raise ValueError(f"File {file_path} does not exist.")

        return pd.read_csv(file_path)

    if dataset is None:
        return _list_datasets()

    raise ValueError(
        f"Dataset {dataset} not found! The following are available:\n{_list_datasets()}"
    )


def _list_datasets() -> str:
    """
    Create a string listing all the available datasets.

    The string includes the datasets' names, their paths and descriptions.

    Returns
    -------
    str
        String listing all the available datasets.
    """
    lines = []
    for filename, resource in DATASETS.items():
        file_path = resource.path
        location = (
            "location: file does not exist"
            if not os.path.exists(file_path)
            else f"location: {file_path}"
        )
        lines.append(
            f"{filename}\n{'=' * len(filename)}\n{resource.description}\n{location}"
        )

    return f"\n\n{10 * '-'}\n\n".join(lines)



"""Utility functions for creating pytensor Ops from onnx model files."""

from .onnx import make_jax_logp_funcs_from_onnx, make_jax_logp_ops, make_pytensor_logp
from .onnx2pt import pt_interpret_onnx
from .onnx2xla import interpret_onnx

__all__ = [
    "interpret_onnx",
    "make_jax_logp_funcs_from_onnx",
    "make_jax_logp_ops",
    "make_pytensor_logp",
    "pt_interpret_onnx",
]



# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""An ONNX to XLA compiler by JAX-tracing a Numpy-backed ONNX interpreter.

Tips for extending this file to add more Ops:

First, Look up for ONNX Ops here:
    https://github.com/onnx/onnx/blob/main/docs/Operators.md#SequenceEmpty

If there is an equivalent op in JAX, just extend the mapping below and add a
lambda function with the same number of inputs as the jax op and wrap the result
in a list.

For example, for ONNX Op Tanh, there is a jnp.tanh function. So we just create
a lambda function like this:

    lambda x: [jnp.tanh(x)]

If there is no equivalent in JAX, then we need to write a function. The `onnx_gemm`
function is an example of such a function. The number of argument is the same as
the number of inputs specified in the Opset document above. If there are any optional
inputs, also add these inputs as parameters with default values.
"""

import jax.numpy as jnp
import numpy as np
import onnx
from jax import lax
from onnx import numpy_helper


def _asarray(proto):
    return numpy_helper.to_array(proto).reshape(tuple(proto.dims))


# pylint: disable=E1101
attr_types = dict(onnx.AttributeProto.AttributeType.items())
attribute_handlers = {
    attr_types["FLOAT"]: lambda a: a.f,
    attr_types["INT"]: lambda a: a.i,
    attr_types["STRING"]: lambda a: a.s,
    attr_types["TENSOR"]: lambda a: _asarray(a.t),
    attr_types["FLOATS"]: lambda a: a.floats,
    attr_types["INTS"]: lambda a: a.ints,
    attr_types["STRINGS"]: lambda a: a.strings,
    attr_types["TENSORS"]: lambda a: [_asarray(x) for x in a.tensors],
}


def onnx_maxpool(x, kernel_shape, pads=None, strides=None):
    """Numpy-backed implementation of ONNX MaxPool op."""
    prefix = (1,) * (x.ndim - len(kernel_shape))
    dims = prefix + tuple(kernel_shape)
    pads = tuple(pads) if pads else [0] * len(kernel_shape)
    strides = (prefix + tuple(strides)) if strides else [1] * len(kernel_shape)
    return [lax.reduce_window(x, -jnp.inf, lax.max, dims, strides, "VALID")]


def onnx_conv(
    x,
    w,
    b=0,
    group=1,
    kernel_shape=None,
    pads=None,
    strides=None,
    dilations=None,
    auto_pad=None,
):
    """Numpy-backed implementation of ONNX Conv op."""
    assert group == 1
    kernel_shape = kernel_shape or w.shape
    strides = strides or [1] * (w.ndim - 2)
    if auto_pad:
        auto_pad = "SAME" if auto_pad.startswith(b"SAME") else "VALID"
        pads = lax.padtype_to_pads(x.shape[2:], w.shape[2:], strides, auto_pad)
    else:
        pads = pads or [0] * (w.ndim - 2)
    lhs_dilation = [1] * (w.ndim - 2)
    rhs_dilation = dilations or [1] * (w.ndim - 2)
    return [
        lax.conv_with_general_padding(x, w, strides, pads, lhs_dilation, rhs_dilation)
        + b
    ]


def onnx_add(a, b, axis=None, broadcast=True):
    """Numpy-backed implementation of ONNX Add op."""
    if broadcast:
        axis = (a.dim - b.ndim) if axis is None else axis % a.ndim
        assert a.shape[axis:][: b.ndim] == b.shape
        b_shape = np.ones(a.ndim, dtype="int64")
        b_shape[axis : axis + b.ndim] = b.shape
        b = jnp.reshape(b, b_shape)
    return [a + b]


# Added by HSSM Developers
def onnx_gemm(
    a, b, c=0.0, alpha=1.0, beta=1.0, transA=0, transB=0
):  # pylint: disable=C0103
    """Numpy-backed implementatio of Onnx Gemm op."""
    a = jnp.transpose(a) if transA else a
    b = jnp.transpose(b) if transB else b
    # jax.debug.print("a: {}", a.dtype)
    # jax.debug.print("b: {}", b.dtype)

    return [alpha * jnp.matmul(a, b) + beta * c]


onnx_ops = {
    "Add": onnx_add,
    "Constant": lambda value: [value],
    "Conv": onnx_conv,
    "MatMul": lambda x, y: [jnp.matmul(x, y)],
    "MaxPool": onnx_maxpool,
    "Relu": lambda x: [jnp.maximum(x, 0)],
    "Reshape": lambda x, shape: [jnp.reshape(x, shape)],
    # Added by HSSM developers
    "Tanh": lambda x: [jnp.tanh(x)],
    "Gemm": onnx_gemm,
}


def interpret_onnx(graph, *args):
    """Transform model in onnx to pytensor.

    Parameters
    ----------
    graph
        The computation graph.
    args
        Inputs to the graph.

    Returns
    -------
        The result of the computation.
    """
    vals = dict(
        {n.name: a for n, a in zip(graph.input, args)},
        **{n.name: _asarray(n) for n in graph.initializer}
    )
    for node in graph.node:
        args = (vals[name] for name in node.input)
        attrs = {a.name: attribute_handlers[a.type](a) for a in node.attribute}
        outputs = onnx_ops[node.op_type](*args, **attrs)
        for name, output in zip(node.output, outputs):
            vals[name] = output
    return [vals[n.name] for n in graph.output]



"""pytensor implementation of the Wiener First Passage Time Distribution.

This code is based on Sam Mathias's Pytensor/Theano implementation
of the WFPT distribution here:
https://gist.github.com/sammosummo/c1be633a74937efaca5215da776f194b.
"""

from __future__ import annotations

from typing import Type

import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt
from numpy import inf
from pymc.distributions.dist_math import check_parameters

from ..distribution_utils.dist import make_distribution

LOGP_LB = pm.floatX(-66.1)


def k_small(rt: np.ndarray, err: float) -> np.ndarray:
    """Determine number of terms needed for small-t expansion.

    Parameters
    ----------
    rt
        A 1D numpy array of flipped R.... T.....s. (0, inf).
    err
        Error bound.

    Returns
    -------
    np.ndarray
        A 1D at array of k_small.
    """
    ks = 2 + pt.sqrt(-2 * rt * pt.log(2 * np.sqrt(2 * np.pi * rt) * err))
    ks = pt.max(pt.stack([ks, pt.sqrt(rt) + 1]), axis=0)
    ks = pt.switch(2 * pt.sqrt(2 * np.pi * rt) * err < 1, ks, 2)

    return ks


def k_large(rt: np.ndarray, err: float) -> np.ndarray:
    """Determine number of terms needed for large-t expansion.

    Parameters
    ----------
    rt
        An 1D numpy array of flipped RTs. (0, inf).
    err
        Error bound

    Returns
    -------
    np.ndarray
        A 1D at array of k_large.
    """
    kl = pt.sqrt(-2 * pt.log(np.pi * rt * err) / (np.pi**2 * rt))
    kl = pt.max(pt.stack([kl, 1.0 / (np.pi * pt.sqrt(rt))]), axis=0)
    kl = pt.switch(np.pi * rt * err < 1, kl, 1.0 / (np.pi * pt.sqrt(rt)))

    return kl


def compare_k(rt: np.ndarray, err: float) -> np.ndarray:
    """Compute and compare k_small with k_large.

    Parameters
    ----------
    rt
        An 1D numpy of flipped RTs. (0, inf).
    err
        Error bound.

    Returns
    -------
    np.ndarray
        A 1D boolean at array of which implementation should be used.
    """
    ks = k_small(rt, err)
    kl = k_large(rt, err)

    return ks < kl


def get_ks(k_terms: int, fast: bool) -> np.ndarray:
    """Return an array of ks.

    Returns an array of ks given the number of terms needed to approximate the sum of
    the infinite series.

    Parameters
    ----------
    k_terms
        number of terms needed
    fast
        whether the function is used in the fast of slow expansion.

    Returns
    -------
    np.ndarray
        An array of ks.
    """
    ks = (
        pt.arange(-pt.floor((k_terms - 1) / 2), pt.ceil((k_terms - 1) / 2) + 1)
        if fast
        else pt.arange(1, k_terms + 1).reshape((-1, 1))
    )

    return ks.astype(pytensor.config.floatX)


def ftt01w_fast(tt: np.ndarray, w: float, k_terms: int) -> np.ndarray:
    """Perform fast computation of ftt01w.

    Density function for lower-bound first-passage times with drift rate set to 0 and
    upper bound set to 1, calculated using the fast-RT expansion.

    Parameters
    ----------
    tt
        Flipped, normalized RTs. (0, inf).
    w
        Normalized decision starting point. (0, 1).
    k_terms
        number of terms to use to approximate the PDF.

    Returns
    -------
    np.ndarray
        The approximated function f(tt|0, 1, w).
    """
    # Slightly changed the original code to mimic the paper and
    # ensure correctness
    k = get_ks(k_terms, fast=True)

    # A log-sum-exp trick is used here
    y = w + 2 * k.reshape((-1, 1))
    r = -pt.power(y, 2) / (2 * tt)
    c = pt.max(r, axis=0)
    p = pt.exp(c) * pt.sum(y * pt.exp(r - c), axis=0)
    # Normalize p
    p = p / pt.sqrt(2 * np.pi * pt.power(tt, 3))

    return p


def ftt01w_slow(tt: np.ndarray, w: float, k_terms: int) -> np.ndarray:
    """Perform slow computation of ftt01w.

    Density function for lower-bound first-passage times with drift rate set to 0 and
    upper bound set to 1, calculated using the slow-RT expansion.

    Parameters
    ----------
    tt
        Flipped, normalized RTs. (0, inf).
    w
        Normalized decision starting point. (0, 1).
    k_terms
        number of terms to use to approximate the PDF.

    Returns
    -------
    np.ndarray
        The approximated function f(tt|0, 1, w).
    """
    k = get_ks(k_terms, fast=False)
    y = k * pt.sin(k * np.pi * w)
    r = -pt.power(k, 2) * pt.power(np.pi, 2) * tt / 2
    p = pt.sum(y * pt.exp(r), axis=0) * np.pi

    return p


def ftt01w(
    rt: np.ndarray,
    a: float,
    w: float,
    err: float = 1e-7,
    k_terms: int = 10,
) -> np.ndarray:
    """Compute the approximate density of f(tt|0,1,w).

    Parameters
    ----------
    rt
        Flipped Response Rates. (0, inf).
    a
        Value of decision upper bound. (0, inf).
    w
        Normalized decision starting point. (0, 1).
    err
        Error bound.
    k_terms
        number of terms to use to approximate the PDF.

    Returns
    -------
    np.ndarray
        The Approximated density of f(tt|0,1,w).
    """
    tt = rt / a**2.0

    lambda_rt = compare_k(tt, err)

    p_fast = ftt01w_fast(tt, w, k_terms)
    p_slow = ftt01w_slow(tt, w, k_terms)

    p = pt.switch(lambda_rt, p_fast, p_slow)

    return p


def logp_ddm(
    data: np.ndarray,
    v: float,
    a: float,
    z: float,
    t: float,
    err: float = 1e-15,
    k_terms: int = 20,
    epsilon: float = 1e-15,
) -> np.ndarray:
    """Compute analytical likelihood for the DDM model with `sv`.

    Computes the log-likelihood of the drift diffusion model f(t|v,a,z) using
    the method and implementation of Navarro & Fuss, 2009.

    Parameters
    ----------
    data
        data: 2-column numpy array of (response time, response)
    v
        Mean drift rate. (-inf, inf).
    a
        Value of decision upper bound. (0, inf).
    z
        Normalized decision starting point. (0, 1).
    t
        Non-decision time [0, inf).
    err
        Error bound.
    k_terms
        number of terms to use to approximate the PDF.
    epsilon
        A small positive number to prevent division by zero or
        taking the log of zero.

    Returns
    -------
    np.ndarray
        The analytical likelihoods for DDM.
    """
    data = pt.reshape(data, (-1, 2))
    rt = pt.abs(data[:, 0])
    response = data[:, 1]
    flip = response > 0
    a = a * 2.0
    v_flipped = pt.switch(flip, -v, v)  # transform v if x is upper-bound response
    z_flipped = pt.switch(flip, 1 - z, z)  # transform z if x is upper-bound response
    rt = rt - t

    p = pt.maximum(ftt01w(rt, a, z_flipped, err, k_terms), pt.exp(LOGP_LB))

    logp = pt.where(
        rt <= epsilon,
        LOGP_LB,
        pt.log(p)
        - v_flipped * a * z_flipped
        - (v_flipped**2 * rt / 2.0)
        - 2.0 * pt.log(a),
    )

    checked_logp = check_parameters(logp, a >= 0, msg="a >= 0")
    checked_logp = check_parameters(checked_logp, z >= 0, msg="z >= 0")
    checked_logp = check_parameters(checked_logp, z <= 1, msg="z <= 1")
    return checked_logp


def logp_ddm_sdv(
    data: np.ndarray,
    v: float,
    a: float,
    z: float,
    t: float,
    sv: float,
    err: float = 1e-15,
    k_terms: int = 20,
    epsilon: float = 1e-15,
) -> np.ndarray:
    """Compute the log-likelihood of the drift diffusion model f(t|v,a,z).

    Using the method and implementation of Navarro & Fuss, 2009.

    Parameters
    ----------
    data
        2-column numpy array of (response time, response)
    v
        Mean drift rate. (-inf, inf).
    a
        Value of decision upper bound. (0, inf).
    z
        Normalized decision starting point. (0, 1).
    t
        Non-decision time [0, inf).
    sv
        Standard deviation of the drift rate [0, inf).
    err
        Error bound.
    k_terms
        number of terms to use to approximate the PDF.
    epsilon
        A small positive number to prevent division by zero or taking the log of zero.

    Returns
    -------
    np.ndarray
        The log likelihood of the drift diffusion model with the standard deviation
        of sv.
    """
    if sv == 0:
        return logp_ddm(data, v, a, z, t, err, k_terms, epsilon)

    data = pt.reshape(data, (-1, 2))
    rt = pt.abs(data[:, 0])
    response = data[:, 1]
    flip = response > 0
    a = a * 2.0
    v_flipped = pt.switch(flip, -v, v)  # transform v if x is upper-bound response
    z_flipped = pt.switch(flip, 1 - z, z)  # transform z if x is upper-bound response
    rt = rt - t

    p = pt.maximum(ftt01w(rt, a, z_flipped, err, k_terms), pt.exp(LOGP_LB))

    logp = pt.switch(
        rt <= epsilon,
        LOGP_LB,
        pt.log(p)
        + (
            (a * z_flipped * sv) ** 2
            - 2 * a * v_flipped * z_flipped
            - (v_flipped**2) * rt
        )
        / (2 * (sv**2) * rt + 2)
        - 0.5 * pt.log(sv**2 * rt + 1)
        - 2 * pt.log(a),
    )

    checked_logp = check_parameters(logp, a >= 0, msg="a >= 0")
    checked_logp = check_parameters(checked_logp, z >= 0, msg="z >= 0")
    checked_logp = check_parameters(checked_logp, z <= 1, msg="z <= 1")
    checked_logp = check_parameters(checked_logp, sv > 0, msg="sv > 0")
    return checked_logp


ddm_bounds = {
    "v": (-inf, inf),
    "a": (0.0, inf),
    "z": (0.0, 1.0),
    "t": (0.0, inf),
}
ddm_sdv_bounds = ddm_bounds | {"sv": (0.0, inf)}

ddm_params = ["v", "a", "z", "t"]
ddm_sdv_params = ddm_params + ["sv"]

DDM: Type[pm.Distribution] = make_distribution(
    "ddm",
    logp_ddm,
    list_params=["v", "a", "z", "t"],
    bounds=ddm_bounds,
)

DDM_SDV: Type[pm.Distribution] = make_distribution(
    "ddm_sdv",
    logp_ddm_sdv,
    list_params=ddm_sdv_params,
    bounds=ddm_sdv_bounds,
)



"""Likelihood Approximation Network (LAN) utilities.

This module handles LAN-related operations, such as producing log-likelihood functions
from onnx model files and wrapping jax log-likelihood functions in pytensor Ops.
"""

from __future__ import annotations

from os import PathLike
from typing import Callable

import jax.numpy as jnp
import numpy as np
import onnx
import pytensor
import pytensor.tensor as pt
from jax import jit, vjp, vmap
from numpy.typing import ArrayLike
from pytensor.graph import Apply, Op
from pytensor.link.jax.dispatch import jax_funcify

from .onnx2pt import pt_interpret_onnx
from .onnx2xla import interpret_onnx

LogLikeFunc = Callable[..., ArrayLike]
LogLikeGrad = Callable[..., ArrayLike]


def make_jax_logp_funcs_from_onnx(
    model: str | PathLike | onnx.ModelProto,
    params_is_reg: list[bool],
) -> tuple[LogLikeFunc, LogLikeGrad, LogLikeFunc]:
    """Make a jax function and its Vector-Jacobian Product from an ONNX Model.

    Parameters
    ----------
    model:
        A path or url to the ONNX model, or an ONNX Model object that's
        already loaded.
    params_is_reg:
        A list of booleans indicating whether the parameters are regressions.
        Parameters that are regressions will not be vectorized in likelihood
        calculations.

    Returns
    -------
    tuple[LogLikeFunc, LogLikeGrad, LogLikeFunc]
        A triple of jax functions. The first calculates the
        forward pass, the second calculates the VJP, and the third is
        the forward-pass that's not jitted.
    """
    loaded_model = (
        onnx.load(str(model)) if isinstance(model, (str, PathLike)) else model
    )

    def logp(data: np.ndarray, *dist_params: float) -> float:
        """Compute the log-likelihood.

        A function that computes the element-wise log-likelihoods given one data point
        and arbitrary numbers of parameters as scalars.

        Parameters
        ----------
        data
            A 1-D length 2 array with response time and response that
            represents one data point
        dist_params
            A list of parameters used in the likelihood computation.

        Returns
        -------
        float
            The element-wise log-likelihoods.
        """
        # Makes a matrix to feed to the LAN model
        input_vector = jnp.concatenate((jnp.array(dist_params), data))

        return interpret_onnx(loaded_model.graph, input_vector)[0].squeeze()

    # The vectorization of the logp function
    vmap_logp = vmap(
        logp,
        in_axes=[0] + [0 if is_regression else None for is_regression in params_is_reg],
    )

    def vjp_vmap_logp(
        data: np.ndarray, *dist_params: list[float | ArrayLike], gz: ArrayLike
    ) -> list[ArrayLike]:
        """Compute the VJP of the log-likelihood function.

        Parameters
        ----------
        data
            A two-column numpy array with response time and response.
        dist_params
            A list of parameters used in the likelihood computation.
        gz
            The value of vmap_logp at which the VJP is evaluated, typically is just
            vmap_logp(data, *dist_params)

        Returns
        -------
        list[ArrayLike]
            The VJP of the log-likelihood function computed at gz.
        """
        _, vjp_fn = vjp(vmap_logp, data, *dist_params)
        return vjp_fn(gz)[1:]

    return jit(vmap_logp), jit(vjp_vmap_logp), vmap_logp


def make_jax_logp_ops(
    logp: LogLikeFunc,
    logp_vjp: LogLikeGrad,
    logp_nojit: LogLikeFunc,
) -> Op:
    """Wrap the JAX functions and its gradient in pytensor Ops.

    Parameters
    ----------
    logp
        A JAX function that represents the feed-forward operation of the LAN
        network.
    logp_vjp
        The Jax function that calculates the VJP of the logp function.
    logp_nojit
        The non-jit version of logp.

    Returns
    -------
    Op
        An pytensor op that wraps the feed-forward operation and can be used with
        pytensor.grad.
    """

    class LANLogpOp(Op):  # pylint: disable=W0223
        """Wraps a JAX function in an pytensor Op."""

        def make_node(self, data, *dist_params):
            """Take the inputs to the Op and puts them in a list.

            Also specifies the output types in a list, then feed them to the Apply node.

            Parameters
            ----------
            data
                A two-column numpy array with response time and response.
            dist_params
                A list of parameters used in the likelihood computation. The parameters
                can be both scalars and arrays.
            """
            inputs = [
                pt.as_tensor_variable(data),
            ] + [pt.as_tensor_variable(dist_param) for dist_param in dist_params]

            outputs = [pt.vector()]

            return Apply(self, inputs, outputs)

        def perform(self, node, inputs, output_storage):
            """Perform the Apply node.

            Parameters
            ----------
            inputs
                This is a list of data from which the values stored in
                output_storage are to be computed using non-symbolic language.
            output_storage
                This is a list of storage cells where the output
                is to be stored. A storage cell is a one-element list. It is
                forbidden to change the length of the list(s) contained in
                output_storage. There is one storage cell for each output of
                the Op.
            """
            result = logp(*inputs)
            output_storage[0][0] = np.asarray(result, dtype=node.outputs[0].dtype)

        def grad(self, inputs, output_gradients):
            """Perform the pytensor.grad() operation.

            Parameters
            ----------
            inputs
                The same as the inputs produced in `make_node`.
            output_gradients
                Holds the results of the perform `perform` method.

            Notes
            -----
                It should output the VJP of the Op. In other words, if this `Op`
                outputs `y`, and the gradient at `y` is grad(x), the required output
                is y*grad(x).
            """
            results = lan_logp_vjp_op(inputs[0], *inputs[1:], gz=output_gradients[0])
            output = [
                pytensor.gradient.grad_not_implemented(self, 0, inputs[0]),
            ] + results

            return output

    class LANLogpVJPOp(Op):  # pylint: disable=W0223
        """Wraps the VJP operation of a jax function in an pytensor op."""

        def make_node(self, data, *dist_params, gz):
            """Take the inputs to the Op and puts them in a list.

            Also specifies the output types in a list, then feed them to the Apply node.

            Parameters
            ----------
            data:
                A two-column numpy array with response time and response.
            dist_params:
                A list of parameters used in the likelihood computation.
            """
            inputs = (
                [
                    pt.as_tensor_variable(data),
                ]
                + [pt.as_tensor_variable(dist_param) for dist_param in dist_params]
                + [pt.as_tensor_variable(gz)]
            )
            outputs = [inp.type() for inp in inputs[1:-1]]

            return Apply(self, inputs, outputs)

        def perform(self, node, inputs, outputs):
            """Perform the Apply node.

            Parameters
            ----------
            inputs
                This is a list of data from which the values stored in
                `output_storage` are to be computed using non-symbolic language.
            output_storage
                This is a list of storage cells where the output
                is to be stored. A storage cell is a one-element list. It is
                forbidden to change the length of the list(s) contained in
                output_storage. There is one storage cell for each output of
                the Op.
            """
            results = logp_vjp(inputs[0], *inputs[1:-1], gz=inputs[-1])

            for i, result in enumerate(results):
                outputs[i][0] = np.asarray(result, dtype=node.outputs[i].dtype)

    lan_logp_op = LANLogpOp()
    lan_logp_vjp_op = LANLogpVJPOp()

    # Unwraps the JAX function for sampling with JAX backend.
    @jax_funcify.register(LANLogpOp)
    def logp_op_dispatch(op, **kwargs):  # pylint: disable=W0612,W0613
        return logp_nojit

    return lan_logp_op


def make_pytensor_logp(
    model: str | PathLike | onnx.ModelProto,
) -> Callable[..., ArrayLike]:
    """Convert onnx model file to pytensor.

    Parameters
    ----------
    model
        A path or url to the ONNX model, or an ONNX Model object that's
        already loaded.
    params_is_reg:
        A list of booleans indicating whether the parameters are regressions.
        Parameters that are regressions will not be vectorized in likelihood
        calculations.

    Returns
    -------
    Callable
        The logp function that applies the ONNX model to data and returns the element-
        wise log-likelihoods.
    """
    loaded_model: onnx.ModelProto = (
        onnx.load(str(model)) if isinstance(model, (str, PathLike)) else model
    )

    def logp(data: np.ndarray, *dist_params: list[float | ArrayLike]) -> ArrayLike:
        # Specify input layer of MLP
        data = data.reshape((-1, 2))
        inputs = pt.zeros((data.shape[0], len(dist_params) + 2))
        for i, dist_param in enumerate(dist_params):
            inputs = pt.set_subtensor(
                inputs[:, i],
                dist_param,
            )
        inputs = pt.set_subtensor(inputs[:, -2:], data)

        # Returns elementwise log-likelihoods
        return pt.squeeze(pt_interpret_onnx(loaded_model.graph, inputs)[0])

    return logp



"""Utility functions and classes.

HSSM has to reconcile with two representations: it's own representation as an HSSM and
the representation acceptable to Bambi. The two are not equivalent. This file contains
the Param class that reconcile these differences.

The Param class is an abstraction that stores the parameter specifications and turns
these representations in Bambi-compatible formats through convenience function
_parse_bambi().
"""

from __future__ import annotations

import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Iterable, Literal, NewType

import bambi as bmb
import pytensor
from bambi.terms import CommonTerm, GroupSpecificTerm, HSGPTerm, OffsetTerm
from huggingface_hub import hf_hub_download
from jax.config import config
from pymc.model_graph import ModelGraph
from pytensor import function

if TYPE_CHECKING:
    from .param import Param

_logger = logging.getLogger("hssm")

REPO_ID = "franklab/HSSM"


def download_hf(path: str):
    """
    Download a file from a HuggingFace repository.

    Parameters
    ----------
    path : str
        The path of the file to download in the repository.

    Returns
    -------
    str
        The local path where the file is downloaded.

    Notes
    -----
    The repository is specified by the REPO_ID constant,
    which should be a valid HuggingFace.co repository ID.
    The file is downloaded using the HuggingFace Hub's
     hf_hub_download function.
    """
    return hf_hub_download(repo_id=REPO_ID, filename=path)


def merge_dicts(dict1: dict, dict2: dict) -> dict:
    """Recursively merge two dictionaries."""
    merged = deepcopy(dict1)
    for key, value in dict2.items():
        if key in merged and isinstance(merged[key], dict) and isinstance(value, dict):
            merged[key] = merge_dicts(merged[key], value)
        else:
            merged[key] = value
    return merged


def make_alias_dict_from_parent(parent: Param) -> dict[str, str]:
    """Make aliases from the parent parameter.

    From a Param object that represents a parent parameter in Bambi,
    returns a dict that represents how Bambi should alias its parameters to
    make it more HSSM-friendly.

    Parameters
    ----------
        parent: A Param object that represents a parent parameter.

    Returns
    -------
        A dict that indicates how Bambi should alias its parameters.
    """
    assert parent.is_parent, "This Param object should be a parent!"

    result_dict = {"c(rt, response)": "rt,response"}

    # The easy case. We will just alias "Intercept" as the actual name of the
    # parameter
    if not parent.is_regression:
        result_dict |= {"Intercept": parent.name}

        return result_dict

    # The regression case:
    # In this case, the name of the response variable should actually be
    # the name of the parent parameter
    result_dict["c(rt, response)"] = parent.name

    return result_dict


def get_alias_dict(model: bmb.Model, parent: Param) -> dict[str, str | dict]:
    """Make a list of aliases.

    Iterates through a list of Param objects, and aliases a Bambi model's parameters
    to make it more HSSM-friendly.

    Parameters
    ----------
    model
        A Bambi model.
    parent
        The Param representation of the parent parameter.

    Returns
    -------
    dict[str, str | dict]
        A dict that indicates how Bambi should alias its parameters.
    """
    parent_name = parent.name

    if len(model.distributional_components) == 1:
        alias_dict: dict[str, Any] = {"c(rt, response)": "rt,response"}
        if not parent.is_regression:
            alias_dict |= {"Intercept": parent_name}
        else:
            for name, term in model.response_component.terms.items():
                if isinstance(
                    term, (CommonTerm, GroupSpecificTerm, HSGPTerm, OffsetTerm)
                ):
                    alias_dict |= {name: f"{parent_name}_{name}"}
    else:
        alias_dict = {"c(rt, response)": {"c(rt, response)": "rt,response"}}
        for component_name, component in model.distributional_components.items():
            if component.response_kind == "data":
                if not parent.is_regression:
                    alias_dict["c(rt, response)"] |= {"Intercept": parent_name}
                else:
                    for name, term in model.response_component.terms.items():
                        if isinstance(
                            term, (CommonTerm, GroupSpecificTerm, HSGPTerm, OffsetTerm)
                        ):
                            alias_dict["c(rt, response)"] |= {
                                name: f"{parent_name}_{name}"
                            }
            else:
                alias_dict[component_name] = {component_name: component_name}

    for name in model.constant_components.keys():
        alias_dict |= {name: name}

    return alias_dict


def fast_eval(var):
    """Fast evaluation of a variable.

    Notes
    -----
    This is a helper function required for one of the functions below.
    """
    return function([], var, mode="FAST_COMPILE")()


VarName = NewType("VarName", str)


class HSSMModelGraph(ModelGraph):
    """Customize PyMC's ModelGraph class to inject the missing parent parameter.

    Notes
    -----
    This is really a hack. There might be better ways to get around the
    parent parameter issue.
    """

    def __init__(self, model, parent):
        self.parent = parent
        super().__init__(model)

    def make_graph(
        self, var_names: Iterable[VarName] | None = None, formatting: str = "plain"
    ):
        """Make graphviz Digraph of PyMC model.

        Returns
        -------
            graphviz.Digraph

        Notes
        -----
            This is a slightly modified version of the code in:
            https://github.com/pymc-devs/pymc/blob/main/pymc/model_graph.py

            Credit for this code goes to PyMC developers.
        """
        try:
            import graphviz  # pylint: disable=C0415
        except ImportError as e:
            e.msg = (
                "This function requires the python library graphviz, "
                + "along with binaries. "
                + "The easiest way to install all of this is by running\n\n"
                + "\tconda install -c conda-forge python-graphviz"
            )
            raise e
        graph = graphviz.Digraph(self.model.name)
        for plate_label, all_var_names in self.get_plates(var_names).items():
            if plate_label:
                # must be preceded by 'cluster' to get a box around it
                with graph.subgraph(name="cluster" + plate_label) as sub:
                    for var_name in all_var_names:
                        self._make_node(var_name, sub, formatting=formatting)
                    # plate label goes bottom right
                    sub.attr(
                        label=plate_label,
                        labeljust="r",
                        labelloc="b",
                        style="rounded",
                    )

            else:
                for var_name in all_var_names:
                    self._make_node(var_name, graph, formatting=formatting)

        if self.parent.is_regression:
            # Insert the parent parameter that's not included in the graph
            with graph.subgraph(name="cluster" + self.parent.name) as sub:
                sub.node(
                    self.parent.name,
                    label=f"{self.parent.name}\n~\nDeterministic",
                    shape="box",
                )
                shape = fast_eval(self.model["rt,response"].shape)
                plate_label = f"rt,response_obs({shape[0]})"

                sub.attr(
                    label=plate_label,
                    labeljust="r",
                    labelloc="b",
                    style="rounded",
                )

        for child, parents in self.make_compute_graph(var_names=var_names).items():
            # parents is a set of rv names that precede child rv nodes
            for parent in parents:
                if (
                    self.parent.is_regression
                    and parent.startswith(f"{self.parent.name}_")
                    and child == "rt,response"
                ):
                    # Modify the edges so that they point to the
                    # parent parameter
                    graph.edge(parent.replace(":", "&"), self.parent.name)
                else:
                    graph.edge(parent.replace(":", "&"), child.replace(":", "&"))

        if self.parent.is_regression:
            graph.edge(self.parent.name, "rt,response")

        return graph


def set_floatX(dtype: Literal["float32", "float64"], jax: bool = True):
    """Set float types for pytensor and Jax.

    Often we wish to work with a specific type of float in both PyTensor and JAX.
    This function helps set float types in both packages.

    Parameters
    ----------
    dtype
        Either `float32` or `float64`. Float type for pytensor (and jax if `jax=True`).
    jax : optional
        Whether this function also sets float type for JAX by changing the
        `jax_enable_x64` setting in JAX config. Defaults to True.
    """
    if dtype not in ["float32", "float64"]:
        raise ValueError('`dtype` must be either "float32" or "float64".')

    pytensor.config.floatX = dtype
    _logger.info(f"Setting PyTensor floatX type to {dtype}.")

    if jax:
        mapping = dict(float32=False, float64=True)
        config.update("jax_enable_x64", mapping[dtype])

        _logger.info(
            f'Setting "jax_enable_x64" to {mapping[dtype]}. '
            + "If this is not intended, please set `jax` to False."
        )


def _print_prior(term: CommonTerm | GroupSpecificTerm) -> str:
    """Make the output string of a term.

    If prior is a float, print x: prior. Otherwise, print x ~ prior.

    Parameters
    ----------
    term
        A BaseTerm in Bambi

    Returns
    -------
        A string representing the term_name ~ prior pair
    """
    term_name = term.alias or term.name
    prior = term._prior

    if isinstance(prior, float):
        return f"        {term_name}: {prior}"

    return f"        {term_name} ~ {prior}"


def _process_param_in_kwargs(name, prior: float | dict | bmb.Prior) -> dict:
    """Process parameters specified in kwargs.

    Parameters
    ----------
    name
        The name of the parameters.
    prior
        The prior specified.

    Returns
    -------
    dict
        A `dict` that complies with ways to specify parameters in `include`.

    Raises
    ------
    ValueError
        When `prior` is not a `float`, a `dict`, or a `b`mb.Prior` object.
    """
    if isinstance(prior, (int, float, bmb.Prior)):
        return {"name": name, "prior": prior}
    elif isinstance(prior, dict):
        if ("prior" in prior) or ("bounds" in prior):
            return prior | {"name": name}
        else:
            return {"name": name, "prior": prior}
    else:
        raise ValueError(
            f"Parameter {name} must be a float, a dict, or a bmb.Prior object."
        )



"""Utility functions for dynamically building pm.Distributions."""

from ..utils import download_hf
from .dist import (
    make_blackbox_op,
    make_distribution,
    make_distribution_from_blackbox,
    make_distribution_from_onnx,
    make_family,
    make_ssm_rv,
)

__all__ = [
    "download_hf",
    "make_blackbox_op",
    "make_distribution",
    "make_distribution_from_blackbox",
    "make_distribution_from_onnx",
    "make_family",
    "make_ssm_rv",
]



"""Provide default configurations for models in the HSSM class."""
from typing import Any, Literal, Optional

import bambi as bmb

from .likelihoods.analytical import (
    ddm_bounds,
    ddm_params,
    ddm_sdv_params,
    logp_ddm,
    logp_ddm_sdv,
)
from .param import _make_default_prior

SupportedModels = Literal[
    "ddm",
    "ddm_sdv",
    "angle",
    "levy",
    "ornstein",
    "weibull",
    "race_no_bias_angle_4",
    "ddm_seq2_no_bias",
]

LoglikKind = Literal["analytical", "approx_differentiable", "blackbox"]

ConfigParams = Literal[
    "loglik",
    "list_params",
    "default_priors",
    "backend",
    "bounds",
    "rv",
]

Config = dict[ConfigParams, Any]

default_model_config: dict[SupportedModels, dict[Literal[LoglikKind], Config]] = {
    "ddm": {
        "analytical": {
            "loglik": logp_ddm,
            "bounds": ddm_bounds,
            "default_priors": {
                "t": {"name": "Uniform", "lower": 0.0, "upper": 2.0, "initval": 0.1},
            },
        },
        "approx_differentiable": {
            "loglik": "ddm.onnx",
            "bounds": {
                "v": (-3.0, 3.0),
                "a": (0.3, 2.5),
                "z": (0.0, 1.0),
                "t": (0.0, 2.0),
            },
        },
    },
    "ddm_sdv": {
        "analytical": {
            "loglik": logp_ddm_sdv,
            "bounds": ddm_bounds,
            "default_priors": {
                "t": {"name": "Uniform", "lower": 0.0, "upper": 2.0, "initval": 0.1},
            },
        },
        "approx_differentiable": {
            "loglik": "ddm_sdv.onnx",
            "bounds": {
                "v": (-3.0, 3.0),
                "a": (0.3, 2.5),
                "z": (0.1, 0.9),
                "t": (0.0, 2.0),
                "sv": (0.0, 1.0),
            },
        },
    },
    "angle": {
        "approx_differentiable": {
            "loglik": "angle.onnx",
            "bounds": {
                "v": (-3.0, 3.0),
                "a": (0.3, 3.0),
                "z": (0.1, 0.9),
                "t": (0.001, 2.0),
                "theta": (-0.1, 1.3),
            },
        },
    },
    "levy": {
        "approx_differentiable": {
            "loglik": "levy.onnx",
            "bounds": {
                "v": (-3.0, 3.0),
                "a": (0.3, 3.0),
                "z": (0.1, 0.9),
                "alpha": (1.0, 2.0),
                "t": (1e-3, 2.0),
            },
        },
    },
    "ornstein": {
        "approx_differentiable": {
            "loglik": "ornstein.onnx",
            "bounds": {
                "v": (-2.0, 2.0),
                "a": (0.3, 3.0),
                "z": (0.1, 0.9),
                "g": (-1.0, 1.0),
                "t": (1e-3, 2.0),
            },
        },
    },
    "weibull": {
        "approx_differentiable": {
            "loglik": "weibull.onnx",
            "bounds": {
                "v": (-2.5, 2.5),
                "a": (0.3, 2.5),
                "z": (0.2, 0.8),
                "t": (1e-3, 2.0),
                "alpha": (0.31, 4.99),
                "beta": (0.31, 6.99),
            },
        },
    },
    "race_no_bias_angle_4": {
        "approx_differentiable": {
            "loglik": "race_no_bias_angle_4.onnx",
            "bounds": {
                "v0": (0.0, 2.5),
                "v1": (0.0, 2.5),
                "v2": (0.0, 2.5),
                "v3": (0.0, 2.5),
                "a": (1.0, 3.0),
                "z": (0.0, 0.9),
                "ndt": (0.0, 2.0),
                "theta": (-0.1, 1.45),
            },
        },
    },
    "ddm_seq2_no_bias": {
        "approx_differentiable": {
            "loglik": "ddm_seq2_no_bias.onnx",
            "bounds": {
                "vh": (-4.0, 4.0),
                "vl1": (-4.0, 4.0),
                "vl2": (-4.0, 4.0),
                "a": (0.3, 2.5),
                "t": (0.0, 2.0),
            },
        },
    },
}

default_params: dict[SupportedModels, list[str]] = {
    "ddm": ddm_params,
    "ddm_sdv": ddm_sdv_params,
    "angle": ["v", "a", "z", "t", "theta"],
    "levy": ["v", "a", "z", "alpha", "t"],
    "ornstein": ["v", "a", "z", "g", "t"],
    "weibull": ["v", "a", "z", "t", "alpha", "beta"],
    "race_no_bias_angle_4": ["v0", "v1", "v2", "v3", "a", "z", "ndt", "theta"],
    "ddm_seq2_no_bias": ["vh", "vl1", "vl2", "a", "t"],
}


def show_defaults(model: SupportedModels, loglik_kind=Optional[LoglikKind]) -> str:
    """Show the defaults for supported models.

    Parameters
    ----------
    model
        One of the supported model strings.
    loglik_kind : optional
        The kind of likelihood function, by default None, in which case the defaults for
        all likelihoods will be shown.

    Returns
    -------
    str
        A nicely organized printout for the defaults of provided model.
    """
    if model not in default_model_config:
        raise ValueError(f"{model} does not currently have defaults in HSSM.")

    model_config = default_model_config[model]

    output = []
    output.append("Default model config:")
    output.append(f"Model: {model}")

    # Will implement later
    # if "description" in model_config:
    #     output.append("Description:")
    #     output.append(f"    {model_config['description']}")

    output.append(f"Default parameters: {default_params[model]}")
    output.append("")

    if loglik_kind is not None:
        if loglik_kind not in model_config:
            raise ValueError(
                f"{model} does not currently have defaults for `{loglik_kind}` "
                + "log-likelihoods in HSSM."
            )

        output += _show_defaults_helper(model, loglik_kind)

    else:
        for loglik_kind in model_config.keys():
            output += _show_defaults_helper(model, loglik_kind)
            output.append("")

        output = output[:-1]

    return "\r\n".join(output)


def _show_defaults_helper(model: SupportedModels, loglik_kind: LoglikKind) -> list[str]:
    """Show the defaults for supported models.

    Parameters
    ----------
    model
        One of the supported model strings.
    loglik_kind
        The kind of likelihood function.

    Returns
    -------
    list[str]
        A list of nicely organized printout for the defaults of provided model.
    """
    output = []
    params = default_params[model]
    model_defaults = default_model_config[model][loglik_kind]

    output.append(f"Log-likelihood kind: {loglik_kind}")
    output.append(f"Log-likelihood: {model_defaults['loglik']}")
    if loglik_kind == "approx_differentiable":
        output.append("Default backend: jax")
    output.append("Default priors:")

    default_priors = model_defaults.get("default_priors", {})
    default_bounds = model_defaults.get("bounds", {})

    for param in params:
        prior = default_priors.get(param, None)
        if prior is None:
            bounds = default_bounds.get(param, None)
            prior = _make_default_prior(bounds) if bounds is not None else None
        else:
            if isinstance(prior, dict):
                prior = bmb.Prior(**prior)

        output.append(f"    {param} ~ {prior}")

    output.append("Default bounds:")

    for param in params:
        output.append(f"    {param}: {default_bounds.get(param, None)}")

    return output



"""Helper functions for creating a `pymc.Distribution`.

This module provides functions for producing for Wiener First-Passage Time (WFPT)
distributions that support arbitrary log-likelihood functions and random number
generation ops.
"""

from __future__ import annotations

import logging
from os import PathLike
from typing import Any, Callable, Type

import bambi as bmb
import numpy as np
import onnx
import pymc as pm
import pytensor
import pytensor.tensor as pt
from bambi.backend.utils import get_distribution_from_prior
from numpy.typing import ArrayLike
from pytensor.graph.op import Apply, Op
from pytensor.tensor.random.op import RandomVariable
from ssms.basic_simulators import simulator
from ssms.config import model_config as ssms_model_config

from .onnx import make_jax_logp_funcs_from_onnx, make_jax_logp_ops, make_pytensor_logp

LogLikeFunc = Callable[..., ArrayLike]
LogLikeGrad = Callable[..., ArrayLike]

_logger = logging.getLogger("hssm")

OUT_OF_BOUNDS_VAL = pm.floatX(-66.1)


def apply_param_bounds_to_loglik(
    logp: Any,
    list_params: list[str],
    *dist_params: Any,
    bounds: dict[str, tuple[float, float]],
):
    """Adjust the log probability of a model based on parameter boundaries.

    Parameters
    ----------
    logp
        The log probability of the model.
    list_params
        A list of strings representing the names of the distribution parameters.
    dist_params
        The distribution parameters.
    bounds
        Boundaries for parameters in the likelihood.

    Returns
    -------
        The adjusted log likelihoods.
    """
    dist_params_dict = dict(zip(list_params, dist_params))

    bounds = {
        k: (
            pm.floatX(v[0]),
            pm.floatX(v[1]),
        )
        for k, v in bounds.items()
    }

    for param_name, param in dist_params_dict.items():
        # It cannot be assumed that each parameter will have bounds.
        # Skip the paramters that do not have bounds.
        if param_name not in bounds:
            continue

        lower_bound, upper_bound = bounds[param_name]

        out_of_bounds_mask = pt.bitwise_or(
            pt.lt(param, lower_bound), pt.gt(param, upper_bound)
        )

        broadcasted_mask = pt.broadcast_to(
            out_of_bounds_mask, logp.shape
        )  # typing: ignore

        logp = pt.where(broadcasted_mask, OUT_OF_BOUNDS_VAL, logp)

    return logp


def make_ssm_rv(
    model_name: str, list_params: list[str], lapse: bmb.Prior | None = None
) -> Type[RandomVariable]:
    """Build a RandomVariable Op according to the list of parameters.

    Parameters
    ----------
    model_name
        The name of the model. If the `model_name` is not one of the supported
        models in the `ssm_simulators` package, a warning will be raised, and any
        attempt to sample from the RandomVariable will result in an error.
    list_params
        A list of str of all parameters for this `RandomVariable`.
    lapse : optional
        A bmb.Prior object representing the lapse distribution.

    Returns
    -------
    Type[RandomVariable]
        A class of RandomVariable that are to be used in a `pm.Distribution`.
    """
    if model_name not in ssms_model_config:
        _logger.warning(
            "You supplied a model '%s', which is currently not supported in "
            + "the ssm_simulators package. An error will be thrown when sampling from "
            + "the random variable or when using any "
            + "posterior or prior predictive sampling methods.",
            model_name,
        )

    if lapse is not None and list_params[-1] != "p_outlier":
        list_params.append("p_outlier")

    # pylint: disable=W0511, R0903
    class SSMRandomVariable(RandomVariable):
        """SSM random variable."""

        name: str = "SSM_RV"
        # to get around the support checking in PyMC that would result in error
        ndim_supp: int = 1

        ndims_params: list[int] = [0 for _ in list_params]
        dtype: str = "floatX"
        _print_name: tuple[str, str] = ("SSM", "\\operatorname{SSM}")
        _list_params = list_params
        _lapse = lapse

        # PyTensor, as of version 2.12, enforces a check to ensure that
        # at least one parameter has the same ndims as the support.
        # This overrides that check and ensures that the dimension checks are correct.
        # For more information, see this issue
        # https://github.com/lnccbrown/HSSM/issues/36
        def _supp_shape_from_params(*args, **kwargs):
            return (2,)

        # pylint: disable=arguments-renamed,bad-option-value,W0221
        # NOTE: `rng` now is a np.random.Generator instead of RandomState
        # since the latter is now deprecated from numpy
        @classmethod
        def rng_fn(
            cls,
            rng: np.random.Generator,
            *args,
            **kwargs,
        ) -> np.ndarray:
            """Generate random variables from this distribution.

            Parameters
            ----------
            rng
                A `np.random.Generator` object for random state.
            args
                Unnamed arguments of parameters, in the order of `_list_params`, plus
                the last one as size.
            kwargs
                Other keyword arguments passed to the ssms simulator.

            Returns
            -------
            np.ndarray
                An array of `(rt, response)` generated from the distribution.

            Note
            ----
            How size is handled in this method:

            We apply multiple tricks to get this method to work with ssm_simulators.

            First, size could be an array with one element. We squeeze the array and
            use that element as size.

            Then, size could depend on whether the parameters passed to this method.
            If all parameters passed are scalar, that is the easy case. We just
            assemble all parameters into a 1D array and pass it to the `theta`
            argument. In this case, size is number of observations.

            If one of the parameters is a vector, which happens one or more parameters
            is the target of a regression. In this case, we take the size of the
            parameter with the largest size. If size is None, we will set size to be
            this largest size. If size is not None, we check if size is a multiple of
            the largest size. If not, an error is thrown. Otherwise, we assemble the
            parameter as a matrix and pass it as the `theta` argument. The multiple then
            becomes how many samples we draw from each trial.
            """
            # First figure out what the size specified here is
            # Since the number of unnamed arguments is undetermined,
            # we are going to use this hack.
            if "size" in kwargs:
                size = kwargs.pop("size")
            else:
                size = args[-1]
                args = args[:-1]

            # Although we got around the ndims_supp issue, the size parameter passed
            # here is still an array with one element. We need to take it out.
            if not np.isscalar(size):
                size = np.squeeze(size)

            arg_arrays = [np.asarray(arg) for arg in args]

            p_outlier = None

            if cls._list_params[-1] == "p_outlier":
                p_outlier = np.squeeze(arg_arrays.pop(-1))

            iinfo32 = np.iinfo(np.uint32)
            seed = rng.integers(0, iinfo32.max, dtype=np.uint32)

            params = (
                cls._list_params[:-1]
                if cls._list_params[-1] == "p_outlier"
                else cls._list_params
            )

            if params != ssms_model_config[model_name]["params"]:
                raise ValueError(
                    f"The list of parameters in `list_params` {params} "
                    + "is different from the model config in SSM Simulators "
                    + f"({ssms_model_config[model_name]['params']})."
                )

            is_all_scalar = all(arg.size == 1 for arg in arg_arrays)

            if is_all_scalar:
                # All parameters are scalars

                theta = np.stack(arg_arrays)
                n_samples = 1 if not size else size
            else:
                # Preprocess all parameters, reshape them into a matrix of dimension
                # (size, n_params) where size is the number of elements in the largest
                # of all parameters passed to *arg

                elem_max_size = np.argmax([arg.size for arg in arg_arrays])
                max_shape = arg_arrays[elem_max_size].shape

                new_data_size = max_shape[-1]

                theta = np.column_stack(
                    [np.broadcast_to(arg, max_shape).reshape(-1) for arg in arg_arrays]
                )

                if size is None:
                    n_samples = 1
                elif size % new_data_size != 0:
                    raise ValueError(
                        "`size` needs to be a multiple of the size of data"
                    )
                else:
                    n_samples = size // new_data_size

            sim_out = simulator(
                theta=theta,
                model=model_name,
                n_samples=n_samples,
                random_state=seed,
                **kwargs,
            )

            sims_out = np.column_stack([sim_out["rts"], sim_out["choices"]])

            if not is_all_scalar:
                sims_out = sims_out.reshape(
                    (*max_shape[:-1], max_shape[-1] * n_samples, 2)
                )

            if p_outlier is not None:
                assert cls._lapse is not None, (
                    "You have specified `p_outlier`, the probability of the lapse "
                    + "distribution but did not specify the distribution."
                )
                out_shape = sims_out.shape[:-1]
                replace = rng.binomial(n=1, p=p_outlier, size=out_shape).astype(bool)
                replace = np.stack([replace, replace], axis=-1)
                n_draws = np.prod(out_shape)
                lapse_rt = pm.draw(
                    get_distribution_from_prior(cls._lapse).dist(**cls._lapse.args),
                    n_draws,
                    random_seed=rng,
                ).reshape(out_shape)
                lapse_response = rng.binomial(n=1, p=0.5, size=out_shape)
                lapse_output = np.stack(
                    [lapse_rt, lapse_response],
                    axis=-1,
                )
                np.putmask(sims_out, replace, lapse_output)

            return sims_out

    return SSMRandomVariable


def make_distribution(
    rv: str | Type[RandomVariable],
    loglik: LogLikeFunc | pytensor.graph.Op,
    list_params: list[str],
    bounds: dict | None = None,
    lapse: bmb.Prior | None = None,
) -> Type[pm.Distribution]:
    """Make a `pymc.Distribution`.

    Constructs a `pymc.Distribution` from a log-likelihood function and a
    RandomVariable op.

    Parameters
    ----------
    rv
        A RandomVariable Op (a class, not an instance) or a string indicating the model.
        If a string, a RandomVariable class will be created automatically with its
        `rng_fn` class method generated using the simulator identified with this string
        from the `ssm_simulators` package. If the string is not one of the supported
        models in the `ssm_simulators` package, a warning will be raised, and any
        attempt to sample from the RandomVariable will result in an error.
    loglik
        A loglikelihood function. It can be any Callable in Python.
    list_params
        A list of parameters that the log-likelihood accepts. The order of the
        parameters in the list will determine the order in which the parameters
        are passed to the log-likelihood function.
    bounds : optional
        A dictionary with parameters as keys (a string) and its boundaries as values.
        Example: {"parameter": (lower_boundary, upper_boundary)}.
    lapse : optional
        A bmb.Prior object representing the lapse distribution.

    Returns
    -------
    Type[pm.Distribution]
        A pymc.Distribution that uses the log-likelihood function.
    """
    random_variable = make_ssm_rv(rv, list_params, lapse) if isinstance(rv, str) else rv

    if lapse is not None:
        if list_params[-1] != "p_outlier":
            list_params.append("p_outlier")

        data = pt.dvector()
        lapse_logp = pm.logp(
            get_distribution_from_prior(lapse).dist(**lapse.args),
            data,
        )
        lapse_func = pytensor.function(
            [data],
            lapse_logp,
        )

    class SSMDistribution(pm.Distribution):
        """Wiener first-passage time (WFPT) log-likelihood for LANs."""

        # This is just a placeholder because pm.Distribution requires an rv_op
        # Might be updated in the future once

        # NOTE: rv_op is an INSTANCE of RandomVariable
        rv_op = random_variable()
        params = list_params

        @classmethod
        def dist(cls, **kwargs):  # pylint: disable=arguments-renamed
            dist_params = [
                pt.as_tensor_variable(pm.floatX(kwargs[param])) for param in cls.params
            ]
            other_kwargs = {k: v for k, v in kwargs.items() if k not in cls.params}
            return super().dist(dist_params, **other_kwargs)

        def logp(data, *dist_params):  # pylint: disable=E0213
            if list_params[-1] == "p_outlier":
                p_outlier = dist_params[-1]
                dist_params = dist_params[:-1]
                lapse_logp = lapse_func(data[:, 0].eval())

                logp = loglik(data, *dist_params)
                logp = pt.log(
                    (1.0 - p_outlier) * pt.exp(logp)
                    + p_outlier * pt.exp(lapse_logp)
                    + 1e-29
                )
            else:
                logp = loglik(data, *dist_params)

            if bounds is None:
                return logp

            return apply_param_bounds_to_loglik(
                logp, list_params, *dist_params, bounds=bounds
            )

    return SSMDistribution


def make_distribution_from_onnx(
    rv: str | Type[RandomVariable],
    list_params: list[str],
    onnx_model: str | PathLike | onnx.ModelProto,
    backend: str = "jax",
    bounds: dict | None = None,
    params_is_reg: list[bool] | None = None,
    lapse: bmb.Prior | None = None,
) -> Type[pm.Distribution]:
    """Make a PyMC distribution from an ONNX model.

    Produces a PyMC distribution that uses the provided base or ONNX model as
    its log-likelihood function.

    Parameters
    ----------
    rv
        A RandomVariable Op (a class, not an instance) or a string indicating the model.
        If a string, a RandomVariable class will be created automatically with its
        `rng_fn` class method generated using the simulator identified with this string
        from the `ssm_simulators` package. If the string is not one of the supported
        models in the `ssm_simulators` package, a warning will be raised, and any
        attempt to sample from the RandomVariable will result in an error.
    list_params
        A list of the names of the parameters following the order of how they are fed
        to the network.
    onnx_model
        The path of the ONNX model, or one already loaded in memory.
    backend
        Whether to use "pytensor" or "jax" as the backend of the log-likelihood
        computation. If `jax`, the function will be wrapped in an pytensor Op.
    bounds : optional
        A dictionary with parameters as keys (a string) and its boundaries
        as values.Example: {"parameter": (lower_boundary, upper_boundary)}.
    params_is_reg : optional
        A list of booleans indicating whether each parameter in the
        corresponding position in `list_params` is a regression.
    lapse : optional
        A bmb.Prior object representing the lapse distribution.

    Returns
    -------
    Type[pm.Distribution]
        A PyMC Distribution class that uses the ONNX model as its log-likelihood
        function.
    """
    if isinstance(onnx_model, (str, PathLike)):
        onnx_model = onnx.load(str(onnx_model))
    if backend == "pytensor":
        lan_logp_pt = make_pytensor_logp(onnx_model)
        return make_distribution(
            rv,
            lan_logp_pt,
            list_params,
            bounds=bounds,
            lapse=lapse,
        )
    if backend == "jax":
        if params_is_reg is None:
            params_is_reg = [False for param in list_params if param != "p_outlier"]
        logp, logp_grad, logp_nojit = make_jax_logp_funcs_from_onnx(
            onnx_model,
            params_is_reg,
        )
        lan_logp_jax = make_jax_logp_ops(logp, logp_grad, logp_nojit)
        return make_distribution(
            rv,
            lan_logp_jax,
            list_params,
            bounds=bounds,
            lapse=lapse,
        )

    raise ValueError("Currently only 'pytensor' and 'jax' backends are supported.")


def make_family(
    dist: Type[pm.Distribution],
    list_params: list[str],
    link: str | dict[str, bmb.families.Link],
    parent: str = "v",
    likelihood_name: str = "SSM Likelihood",
    family_name="SSM Family",
) -> bmb.Family:
    """Build a family in bambi.

    Parameters
    ----------
    dist
        A pm.Distribution class (not an instance).
    list_params
        A list of parameters for the likelihood function.
    link
        A link function or a dictionary of parameter: link functions.
    parent
        The parent parameter of the likelihood function. Defaults to v.
    likelihood_name
        the name of the likelihood function. Defaults to "SSM Likelihood".
    family_name
        the name of the family. Defaults to "SSM Family".

    Returns
    -------
    bmb.Family
        An instance of a bambi family.
    """
    likelihood = bmb.Likelihood(
        likelihood_name, parent=parent, params=list_params, dist=dist
    )

    family = SSMFamily(family_name, likelihood=likelihood, link=link)

    return family


class SSMFamily(bmb.Family):
    """Extends bmb.Family to get around the dimensionality mismatch."""

    def create_extra_pps_coord(self):
        """Create an extra dimension."""
        return np.arange(2)


def make_blackbox_op(logp: Callable) -> Op:
    """Wrap an arbitrary function in a pytensor Op.

    Parameters
    ----------
    logp
        A python function that represents the log-likelihood function. The function
        needs to have signature of logp(data, *dist_params) where `data` is a
        two-column numpy array and `dist_params`represents all parameters passed to the
        function.

    Returns
    -------
    Op
        An pytensor op that wraps the log-likelihood function.
    """

    class BlackBoxOp(Op):  # pylint: disable=W0223
        """Wraps an arbitrary function in a pytensor Op."""

        def make_node(self, data, *dist_params):
            """Take the inputs to the Op and puts them in a list.

            Also specifies the output types in a list, then feed them to the Apply node.

            Parameters
            ----------
            data
                A two-column numpy array with response time and response.
            dist_params
                A list of parameters used in the likelihood computation. The parameters
                can be both scalars and arrays.
            """
            inputs = [
                pt.as_tensor_variable(data),
            ] + [pt.as_tensor_variable(dist_param) for dist_param in dist_params]

            outputs = [pt.vector()]

            return Apply(self, inputs, outputs)

        def perform(self, node, inputs, output_storage):
            """Perform the Apply node.

            Parameters
            ----------
            inputs
                This is a list of data from which the values stored in
                output_storage are to be computed using non-symbolic language.
            output_storage
                This is a list of storage cells where the output
                is to be stored. A storage cell is a one-element list. It is
                forbidden to change the length of the list(s) contained in
                output_storage. There is one storage cell for each output of
                the Op.
            """
            result = logp(*inputs)
            output_storage[0][0] = np.asarray(result, dtype=node.outputs[0].dtype)

    blackbox_op = BlackBoxOp()
    return blackbox_op


def make_distribution_from_blackbox(
    rv: str | Type[RandomVariable],
    loglik: Callable,
    list_params: list[str],
    bounds: dict | None = None,
) -> Type[pm.Distribution]:
    """Make a `pymc.Distribution`.

    Constructs a `pymc.Distribution` from a blackbox likelihood function

    Parameters
    ----------
    rv
        A RandomVariable Op (a class, not an instance) or a string indicating the model.
        If a string, a RandomVariable class will be created automatically with its
        `rng_fn` class method generated using the simulator identified with this string
        from the `ssm_simulators` package. If the string is not one of the supported
        models in the `ssm_simulators` package, a warning will be raised, and any
        attempt to sample from the RandomVariable will result in an error.
    loglik
        A loglikelihood function. It can be any Callable in Python.
    list_params
        A list of parameters that the log-likelihood accepts. The order of the
        parameters in the list will determine the order in which the parameters
        are passed to the log-likelihood function.
    bounds : optional
        A dictionary with parameters as keys (a string) and its boundaries as values.
        Example: {"parameter": (lower_boundary, upper_boundary)}.

    Returns
    -------
    Type[pm.Distribution]
        A pymc.Distribution that uses the log-likelihood function.
    """
    blackbox_op = make_blackbox_op(loglik)

    return make_distribution(rv, blackbox_op, list_params, bounds)



"""HSSM: Hierarchical Sequential Sampling Models.

A package based on pymc and bambi to perform Bayesian inference for hierarchical
sequential sampling models.

This file defines the entry class HSSM.
"""

from __future__ import annotations

import logging
from copy import deepcopy
from inspect import isclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal, cast

import bambi as bmb
import pymc as pm
from bambi.model_components import DistributionalComponent
from numpy.typing import ArrayLike
from pytensor.graph.op import Op

from hssm.config import (
    Config,
    LoglikKind,
    SupportedModels,
    default_model_config,
    default_params,
)
from hssm.distribution_utils import (
    make_blackbox_op,
    make_distribution,
    make_distribution_from_onnx,
    make_family,
)
from hssm.param import (
    Param,
    _parse_bambi,
)
from hssm.utils import (
    HSSMModelGraph,
    _print_prior,
    _process_param_in_kwargs,
    download_hf,
    get_alias_dict,
    merge_dicts,
)

if TYPE_CHECKING:
    from os import PathLike

    import arviz as az
    import numpy as np
    import pandas as pd
    import pytensor

LogLikeFunc = Callable[..., ArrayLike]

_logger = logging.getLogger("hssm")


class HSSM:
    """The Hierarchical Sequential Sampling Model (HSSM) class.

    Parameters
    ----------
    data
        A pandas DataFrame with the minimum requirements of containing the data with the
        columns "rt" and "response".
    model
        The name of the model to use. Currently supported models are "ddm", "ddm_sdv",
        "angle", "levy", "ornstein", "weibull", "race_no_bias_angle_4",
        "ddm_seq2_no_bias". If any other string is passed, the model will be considered
        custom, in which case all `model_config`, `loglik`, and `loglik_kind` have to be
        provided by the user.
    include : optional
        A list of dictionaries specifying parameter specifications to include in the
        model. If left unspecified, defaults will be used for all parameter
        specifications. Defaults to None.
    model_config : optional
        A dictionary containing the model configuration information. If None is
        provided, defaults will be used if there are any. Defaults to None.
        Fields for this `dict` are usually:

        - `"list_params"`: a list of parameters indicating the parameters of the model.
            The order in which the parameters are specified in this list is important.
            Values for each parameter will be passed to the likelihood function in this
            order.
        - `"backend"`: Only used when `loglik_kind` is `approx_differentiable` and
            an onnx file is supplied for the likelihood approximation network (LAN).
            Valid values are `"jax"` or `"pytensor"`. It determines whether the LAN in
            ONNX should be converted to `"jax"` or `"pytensor"`. If not provided,
            `jax` will be used for maximum performance.
        - `"default_priors"`: A `dict` indicating the default priors for each parameter.
        - `"bounds"`: A `dict` indicating the boundaries for each parameter. In the case
            of LAN, these bounds are training boundaries.
        - `"rv"`: Optional. Can be a `RandomVariable` class containing the user's own
            `rng_fn` function for sampling from the distribution that the user is
            supplying. If not supplied, HSSM will automatically generate a
            `RandomVariable` using the simulator identified by `model` from the
            `ssm_simulators` package. If `model` is not supported in `ssm_simulators`,
            a warning will be raised letting the user know that sampling from the
            `RandomVariable` will result in errors.
    loglik : optional
        A likelihood function. Defaults to None. Requirements are:

        1. if `loglik_kind` is `"analytical"` or `"blackbox"`, a pm.Distribution, a
           pytensor Op, or a Python callable can be used. Signatures are:
            - `pm.Distribution`: needs to have parameters specified exactly as listed in
            `list_params`
            - `pytensor.graph.Op` and `Callable`: needs to accept the parameters
            specified exactly as listed in `list_params`
        2. If `loglik_kind` is `"approx_differentiable"`, then in addition to the
            specifications above, a `str` or `Pathlike` can also be used to specify a
            path to an `onnx` file. If a `str` is provided, HSSM will first look locally
            for an `onnx` file. If that is not successful, HSSM will try to download
            that `onnx` file from Hugging Face hub.
        3. It can also be `None`, in which case a default likelihood function will be
            used
    loglik_kind : optional
        A string that specifies the kind of log-likelihood function specified with
        `loglik`. Defaults to `None`. Can be one of the following:

        - `"analytical"`: an analytical (approximation) likelihood function. It is
            differentiable and can be used with samplers that requires differentiation.
        - `"approx_differentiable"`: a likelihood approximation network (LAN) likelihood
            function. It is differentiable and can be used with samplers that requires
            differentiation.
        - `"blackbox"`: a black box likelihood function. It is typically NOT
            differentiable.
        - `None`, in which a default will be used. For `ddm` type of models, the default
            will be `analytical`. For other models supported, it will be
            `approx_differentiable`. If the model is a custom one, a ValueError
            will be raised.
    p_outlier : optional
        The fixed lapse probability or the prior distribution of the lapse probability.
        Defaults to a fixed value of 0.05. When `None`, the lapse probability will not
        be included in estimation.
    lapse : optional
        The lapse distribution. This argument is required only if `p_outlier` is not
        `None`. Defaults to Uniform(0.0, 10.0).
    hierarchical : optional
        If True, and if there is a `participant_id` field in `data`, will by default
        turn any unspecified parameter theta into a regression with
        "theta ~ 1 + (1|participant_id)" and default priors set by `bambi`.
    **kwargs
        Additional arguments passed to the `bmb.Model` object.

    Attributes
    ----------
    data
        A pandas DataFrame with at least two columns of "rt" and "response" indicating
        the response time and responses.
    list_params
        The list of strs of parameter names.
    model_name
        The name of the model.
    loglik:
        The likelihood function or a path to an onnx file.
    loglik_kind:
        The kind of likelihood used.
    model_config
        A dictionary representing the model configuration.
    model_distribution
        The likelihood function of the model in the form of a pm.Distribution subclass.
    family
        A Bambi family object.
    priors
        A dictionary containing the prior distribution of parameters.
    formula
        A string representing the model formula.
    link
        A string or a dictionary representing the link functions for all parameters.
    params
        A list of Param objects representing model parameters.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        model: SupportedModels | str = "ddm",
        include: list[dict] | None = None,
        model_config: Config | None = None,
        loglik: str | PathLike | LogLikeFunc | pytensor.graph.Op | None = None,
        loglik_kind: LoglikKind | None = None,
        p_outlier: float | dict | bmb.Prior | None = 0.05,
        lapse: dict | bmb.Prior | None = bmb.Prior("Uniform", lower=0.0, upper=10.0),
        hierarchical: bool = True,
        **kwargs,
    ):
        self.data = data
        self._inference_obj = None
        self.hierarchical = hierarchical and "participant_id" in data.columns
        self.has_lapse = p_outlier is not None and p_outlier != 0

        if loglik_kind is None:
            if model not in default_model_config:
                raise ValueError(
                    "When using a custom model, please provide a `loglik_kind.`"
                )
            # Setting loglik_kind to be the first of analytical or
            # approx_differentiable
            for kind in ["analytical", "approx_differentiable", "blackbox"]:
                model = cast(SupportedModels, model)
                if kind in default_model_config[model]:
                    kind = cast(LoglikKind, kind)
                    loglik_kind = kind
                    break
            if loglik_kind is None:
                raise ValueError(
                    "No default model_config is found. Please provide a `loglik_kind."
                )
        else:
            if loglik_kind not in [
                "analytical",
                "approx_differentiable",
                "blackbox",
            ]:
                raise ValueError(
                    "'loglike_kind', when provided, must be one of "
                    + '"analytical", "approx_differentiable", "blackbox".'
                )

        self.loglik_kind = loglik_kind
        self.model_name = model

        # Check if model has default config
        if _model_has_default(self.model_name, self.loglik_kind):
            model = cast(SupportedModels, model)
            default_config = default_model_config[model][loglik_kind]

            self.model_config = (
                deepcopy(default_config)
                if model_config is None
                else merge_dicts(default_config, model_config)
            )
            self.loglik = self.model_config["loglik"] if loglik is None else loglik
            self.list_params = default_params[model][:]
        else:
            # If there is no default, we require a log-likelihood
            if loglik is None:
                raise ValueError("Please provide a valid `loglik`.")
            self.loglik = loglik

            if model not in default_model_config:
                # For custom models, require model_config
                if model_config is None:
                    raise ValueError(
                        "For custom models, please provide a valid `model_config`."
                    )
                if "list_params" not in model_config:
                    raise ValueError(
                        "For custom models, please provide `list_params` in "
                        + "`model_config`."
                    )
                self.model_config = model_config
                self.list_params = model_config["list_params"]
            else:
                # For supported models without configs,
                # We don't require a model_config (because list_params is known)
                model = cast(SupportedModels, model)
                self.model_config = {} if model_config is None else model_config
                self.list_params = default_params[model][:]

        if (
            loglik_kind == "approx_differentiable"
            and "backend" not in self.model_config
        ):
            self.model_config["backend"] = "jax"

        # Logic for determining if p_outlier and lapse is specified correctly.
        # Basically, avoid situations where only one of them is specified.
        self._parent = self.list_params[0]
        if self.has_lapse and lapse is None:
            raise ValueError(
                "You have specified `p_outlier`. Please also specify `lapse`."
            )
        if lapse is not None and not self.has_lapse:
            _logger.warning(
                "You have specified the `lapse` argument to include a lapse "
                + "distribution, but `p_outlier` is set to either 0 or None. "
                + "Your lapse distribution will be ignored."
            )
        if "p_outlier" in self.list_params and self.list_params[-1] != "p_outlier":
            raise ValueError(
                "Please do not include 'p_outlier' in `list_params`. "
                + "We automatically append it to `list_params` when `p_outlier` "
                + "parameter is not None"
            )

        if self.has_lapse and self.list_params[-1] != "p_outlier":
            self.list_params.append("p_outlier")

        if include is None:
            include = []
        params_in_include = [param["name"] for param in include]

        # Process kwargs
        # If any of the keys is found in `list_params` it is a parameter specification
        # We add the parameter specification to `include`, which will be processed later
        # together with other parameter specifications in `include`.
        # Otherwise we create another dict and pass it to `bmb.Model`.
        other_kwargs: dict[Any, Any] = {}
        for k, v in kwargs.items():
            if k in self.list_params:
                if k in params_in_include:
                    raise ValueError(
                        f'Parameter "{k}" is already specified in `include`.'
                    )
                include.append(_process_param_in_kwargs(k, v))
            else:
                other_kwargs |= {k: v}

        # Process p_outliers the same way.
        if self.has_lapse:
            if "p_outlier" in params_in_include:
                raise ValueError(
                    "Please do not specify `p_outlier` in `include`. "
                    + "Please specify it with `p_outlier` instead."
                )
            include.append(_process_param_in_kwargs("p_outlier", p_outlier))

        self.params, self.formula, self.priors, self.link = self._transform_params(
            include, self.model_config
        )

        self._parent_param = self.params[self.list_params[0]]
        assert self._parent_param is not None

        params_is_reg = [
            param.is_regression
            for param_name, param in self.params.items()
            if param_name != "p_outlier"
        ]

        # For parameters that are regression, apply bounds at the likelihood level to
        # ensure that the samples that are out of bounds are discarded (replaced with
        # a large negative value).
        self.bounds = {
            name: param.bounds
            for name, param in self.params.items()
            if param.is_regression and param.bounds is not None
        }

        self.p_outlier = self.params.get("p_outlier")
        self.lapse = lapse if self.has_lapse else None

        ### Logic for different types of likelihoods:
        # -`analytical` and `blackbox`:
        #     loglik should be a `pm.Distribution`` or a Python callable (any arbitrary
        #     function).
        # - `approx_differentiable`:
        #     In addition to `pm.Distribution` and any arbitrary function, it can also
        #     be an str (which we will download from hugging face) or a Pathlike
        #     which we will download and make a distribution.

        # If user has already provided a log-likelihood function as a distribution
        # Use it directly as the distribution
        if isclass(self.loglik) and issubclass(self.loglik, pm.Distribution):
            self.model_distribution = self.loglik
        # If the user has provided an Op
        # Wrap it around with a distribution
        elif isinstance(self.loglik, Op):
            self.model_distribution = make_distribution(
                self.model_config.get("rv", self.model_name),
                loglik=self.loglik,
                list_params=self.list_params,
                bounds=self.bounds,
                lapse=self.lapse,
            )  # type: ignore
        # If the user has provided a callable (an arbitrary likelihood function)
        # If `loglik_kind` is `blackbox`, wrap it in an op and then a distribution
        # Otherwise, we assume that this function is differentiable with `pytensor`
        # and wrap it directly in a distribution.
        elif callable(self.loglik):
            if self.loglik_kind == "blackbox":
                self.loglik = make_blackbox_op(self.loglik)
            self.model_distribution = make_distribution(
                self.model_config.get("rv", self.model_name),
                loglik=self.loglik,
                list_params=self.list_params,
                bounds=self.bounds,
                lapse=self.lapse,
            )  # type: ignore
        # All other situations
        else:
            if self.loglik_kind != "approx_differentiable":
                raise ValueError(
                    "You set `loglik_kind` to `approx_differentiable "
                    + "but did not provide a pm.Distribution, an Op, or a callable "
                    + "as `loglik`."
                )
            if isinstance(self.loglik, str):
                if not Path(self.loglik).exists():
                    self.loglik = download_hf(self.loglik)

            self.model_distribution = make_distribution_from_onnx(
                rv=self.model_config.get("rv", self.model_name),
                onnx_model=self.loglik,
                list_params=self.list_params,
                backend=self.model_config.get("backend", "jax"),
                params_is_reg=params_is_reg,
                bounds=self.bounds,
                lapse=self.lapse,
            )

        self.family = make_family(
            self.model_distribution,
            self.list_params,
            self.link,
            self._parent,
        )

        self.model = bmb.Model(
            self.formula, data, family=self.family, priors=self.priors, **other_kwargs
        )

        self._aliases = get_alias_dict(self.model, self._parent_param)
        self.set_alias(self._aliases)

    def _transform_params(
        self, include: list[dict] | None, model_config: Config
    ) -> tuple[
        dict[str, Param], bmb.Formula, dict | None, dict[str, str | bmb.Link] | str
    ]:
        """Transform parameters.

        Transforms a list of dictionaries containing parameter information into a
        list of Param objects. This function creates a formula, priors,and a link for
        the Bambi package based on the parameters.

        Parameters
        ----------
        include
            A list of dictionaries containing information about the parameters.
        model_config
            A dict for the configuration for the model.

        Returns
        -------
        list[Param], bmb.Formula, dict | None, dict | str
            A tuple of 4 items, the latter 3 are for creating the bambi model.
            - A list of the same length as self.list_params containing Param objects.
            - A bmb.formula object.
            - An optional dict containing prior information for Bambi.
            - An optional dict of link functions for Bambi.
        """
        processed = []
        params: dict[str, Param] = {}
        if include:
            for param_dict in include:
                name = param_dict["name"]
                processed.append(name)
                for k in param_dict.keys():
                    if k not in ["name", "formula", "prior", "link", "bounds"]:
                        raise ValueError(
                            f"Invalid key {k} for the specification of {name}!"
                        )
                param = _create_param(
                    param_dict, model_config, is_parent=name == self._parent
                )
                params[name] = param

        for param_str in self.list_params:
            if param_str not in processed:
                is_parent = param_str == self._parent
                if self.hierarchical:
                    bounds = (
                        model_config["bounds"].get(param_str)
                        if "bounds" in model_config
                        else None
                    )
                    param = Param(
                        param_str,
                        formula="1 + (1|participant_id)",
                        link="identity",
                        bounds=bounds,
                        is_parent=is_parent,
                    )
                else:
                    param = _create_param(param_str, model_config, is_parent=is_parent)
                params[param_str] = param

        sorted_params = {k: params[k] for k in self.list_params}

        return sorted_params, *_parse_bambi(list(sorted_params.values()))

    def sample(
        self,
        sampler: Literal["mcmc", "nuts_numpyro", "nuts_blackjax", "laplace", "vi"]
        | None = None,
        **kwargs,
    ) -> az.InferenceData | pm.Approximation:
        """Perform sampling using the `fit` method via bambi.Model.

        Parameters
        ----------
        sampler
            The sampler to use. Can be one of "mcmc", "nuts_numpyro",
            "nuts_blackjax", "laplace", or "vi". If using `blackbox` likelihoods,
            this cannot be "nuts_numpyro" or "nuts_blackjax". By default it is None, and
            sampler will automatically be chosen: when the model uses the
            `approx_differentiable` likelihood, and `jax` backend, "nuts_numpyro" will
            be used. Otherwise, "mcmc" (the default PyMC NUTS sampler) will be used.
        kwargs
            Other arguments passed to bmb.Model.fit(). Please see [here]
            (https://bambinos.github.io/bambi/api_reference.html#bambi.models.Model.fit)
            for full documentation.

        Returns
        -------
        az.InferenceData | pm.Approximation
            An ArviZ `InferenceData` instance if inference_method is `"mcmc"`
            (default), "nuts_numpyro", "nuts_blackjax" or "laplace". An `Approximation`
            object if `"vi"`.
        """
        if sampler is None:
            if (
                self.loglik_kind == "approx_differentiable"
                and self.model_config.get("backend") == "jax"
            ):
                sampler = "nuts_numpyro"
            else:
                sampler = "mcmc"

        supported_samplers = ["mcmc", "nuts_numpyro", "nuts_blackjax", "laplace", "vi"]

        if sampler not in supported_samplers:
            raise ValueError(
                f"Unsupported sampler '{sampler}', must be one of {supported_samplers}"
            )

        if self.loglik_kind == "blackbox":
            if sampler in ["nuts_blackjax", "nuts_numpyro"]:
                raise ValueError(
                    f"{sampler} sampler does not work with blackbox likelihoods."
                )

            if "step" not in kwargs:
                kwargs["step"] = pm.Slice(model=self.pymc_model)

        if (
            self.loglik_kind == "approx_differentiable"
            and self.model_config.get("backend") == "jax"
            and sampler == "mcmc"
            and kwargs.get("cores", None) != 1
        ):
            _logger.warning(
                "Parallel sampling might not work with `jax` backend and the PyMC NUTS "
                + "sampler on some platforms. Please consider using `nuts_numpyro` or "
                + "`nuts_blackjax` sampler if that is a problem."
            )

        self._inference_obj = self.model.fit(inference_method=sampler, **kwargs)

        return self.traces

    def sample_posterior_predictive(
        self,
        idata: az.InferenceData | None = None,
        data: pd.DataFrame | None = None,
        inplace: bool = True,
        include_group_specific: bool = True,
        kind: Literal["pps", "mean"] = "pps",
    ) -> az.InferenceData | None:
        """Perform posterior predictive sampling from the HSSM model.

        Parameters
        ----------
        idata : optional
            The `InferenceData` object returned by `HSSM.sample()`. If not provided,
            the `InferenceData` from the last time `sample()` is called will be used.
        data : optional
            An optional data frame with values for the predictors that are used to
            obtain out-of-sample predictions. If omitted, the original dataset is used.
        inplace : optional
            If `True` will modify idata in-place and append a `posterior_predictive`
            group to `idata`. Otherwise, it will return a copy of idata with the
            predictions added, by default True.
        include_group_specific : optional
            If `True` will make predictions including the group specific effects.
            Otherwise, predictions are made with common effects only (i.e. group-
            specific are set to zero), by default True.
        kind
            Indicates the type of prediction required. Can be `"mean"` or `"pps"`. The
            first returns draws from the posterior distribution of the mean, while the
            latter returns the draws from the posterior predictive distribution
            (i.e. the posterior probability distribution for a new observation).
            Defaults to `"pps"`.

        Raises
        ------
        ValueError
            If the model has not been sampled yet and idata is not provided.

        Returns
        -------
        az.InferenceData | None
            InferenceData or None
        """
        if idata is None:
            if self._inference_obj is None:
                raise ValueError(
                    "The model has not been sampled yet. "
                    + "Please either provide an idata object or sample the model first."
                )
            idata = self._inference_obj
        return self.model.predict(idata, kind, data, inplace, include_group_specific)

    def sample_prior_predictive(
        self,
        draws: int = 500,
        var_names: str | list[str] | None = None,
        omit_offsets: bool = True,
        random_seed: np.random.Generator | None = None,
    ) -> az.InferenceData:
        """Generate samples from the prior predictive distribution.

        Parameters
        ----------
        draws
            Number of draws to sample from the prior predictive distribution. Defaults
            to 500.
        var_names
            A list of names of variables for which to compute the prior predictive
            distribution. Defaults to ``None`` which means both observed and unobserved
            RVs.
        omit_offsets
            Whether to omit offset terms. Defaults to ``True``.
        random_seed
            Seed for the random number generator.

        Returns
        -------
        az.InferenceData
            ``InferenceData`` object with the groups ``prior``, ``prior_predictive`` and
            ``observed_data``.
        """
        return self.model.prior_predictive(draws, var_names, omit_offsets, random_seed)

    @property
    def pymc_model(self) -> pm.Model:
        """Provide access to the PyMC model.

        Returns
        -------
        pm.Model
            The PyMC model built by bambi
        """
        return self.model.backend.model

    def set_alias(self, aliases: dict[str, str | dict]):
        """Set parameter aliases.

        Sets the aliases according to the dictionary passed to it and rebuild the
        model.

        Parameters
        ----------
        aliases
            A dict specifying the parameter names being aliased and the aliases.
        """
        self.model.set_alias(aliases)
        self.model.build()

    # NOTE: can't annotate return type because the graphviz dependency is
    # optional
    def graph(self, formatting="plain", name=None, figsize=None, dpi=300, fmt="png"):
        """Produce a graphviz Digraph from a built HSSM model.

        Requires graphviz, which may be installed most easily with `conda install -c
        conda-forge python-graphviz`. Alternatively, you may install the `graphviz`
        binaries yourself, and then `pip install graphviz` to get the python bindings.
        See http://graphviz.readthedocs.io/en/stable/manual.html for more information.

        Parameters
        ----------
        formatting
            One of `"plain"` or `"plain_with_params"`. Defaults to `"plain"`.
        name
            Name of the figure to save. Defaults to `None`, no figure is saved.
        figsize
            Maximum width and height of figure in inches. Defaults to `None`, the
            figure size is set automatically. If defined and the drawing is larger than
            the given size, the drawing is uniformly scaled down so that it fits within
            the given size.  Only works if `name` is not `None`.
        dpi
            Point per inch of the figure to save.
            Defaults to 300. Only works if `name` is not `None`.
        fmt
            Format of the figure to save.
            Defaults to `"png"`. Only works if `name` is not `None`.

        Returns
        -------
        graphviz.Graph
            The graph

        Note
        ----
            The code is largely copied from
            https://github.com/bambinos/bambi/blob/main/bambi/models.py
            Credit for the code goes to Bambi developers.
        """
        self.model._check_built()

        graphviz = HSSMModelGraph(
            model=self.pymc_model, parent=self._parent_param
        ).make_graph(formatting=formatting)

        width, height = (None, None) if figsize is None else figsize

        if name is not None:
            graphviz_ = graphviz.copy()
            graphviz_.graph_attr.update(size=f"{width},{height}!")
            graphviz_.graph_attr.update(dpi=str(dpi))
            graphviz_.render(filename=name, format=fmt, cleanup=True)

            return graphviz_

        return graphviz

    def __repr__(self) -> str:
        """Create a representation of the model."""
        output = []

        output.append("Hierarchical Sequential Sampling Model")
        output.append(f"Model: {self.model_name}")
        output.append("")

        output.append("Response variable: rt,response")
        output.append(f"Likelihood: {self.loglik_kind}")
        output.append(f"Observations: {len(self.data)}")
        output.append("")

        output.append("Parameters:")
        output.append("")

        for param in self.params.values():
            if param.name == "p_outlier":
                continue
            name = "c(rt, response)" if param.is_parent else param.name
            output.append(f"{param.name}:")

            component = self.model.components[name]

            # Regression case:
            if param.is_regression:
                assert isinstance(component, DistributionalComponent)
                output.append(f"    Formula: {param.formula}")
                output.append("    Priors:")
                intercept_term = component.intercept_term
                if intercept_term is not None:
                    output.append(_print_prior(intercept_term))
                for _, common_term in component.common_terms.items():
                    output.append(_print_prior(common_term))
                for _, group_specific_term in component.group_specific_terms.items():
                    output.append(_print_prior(group_specific_term))
                output.append(f"    Link: {param.link}")
            # None regression case
            else:
                if param.prior is None:
                    prior = (
                        component.intercept_term.prior
                        if param.is_parent
                        else component.prior
                    )
                else:
                    prior = param._prior if param.is_truncated else param.prior
                output.append(f"    Prior: {prior}")
            output.append(f"    Explicit bounds: {param.bounds}")

        if self.p_outlier is not None:
            # TODO: Allow regression for self.p_outlier
            # Need to determine what the output should look like
            # and whether p should be hierarchical when self.hierarchical is True.
            assert not self.p_outlier.is_regression
            output.append("")
            output.append(f"Lapse probability: {self.p_outlier.prior}")
            output.append(f"Lapse distribution: {self.lapse}")

        return "\r\n".join(output)

    def __str__(self) -> str:
        """Create a string representation of the model."""
        return self.__repr__()

    @property
    def traces(self) -> az.InferenceData | pm.Approximation:
        """Return the trace of the model after sampling.

        Raises
        ------
        ValueError
            If the model has not been sampled yet.

        Returns
        -------
        az.InferenceData | pm.Approximation
            The trace of the model after sampling.
        """
        if not self._inference_obj:
            raise ValueError("Please sample the model first.")

        return self._inference_obj


def _model_has_default(model: SupportedModels | str, loglik_kind: LoglikKind) -> bool:
    """Determine if the specified model has default configs.

    Also checks if `model` and `loglik_kind` are valid.

    Parameters
    ----------
    model
        User-specified model type.
    loglik_kind
        User-specified likelihood kind.

    Returns
    -------
    bool
        Whether the model is supported.
    """
    if model not in default_model_config:
        return False

    model = cast(SupportedModels, model)
    return loglik_kind in default_model_config[model]


def _create_param(param: str | dict, model_config: dict, is_parent: bool) -> Param:
    """Create a Param object.

    Parameters
    ----------
    param
        A dict or str containing parameter settings.
    model_config
        A dict containing the config for the model.
    is_parent
        Indicates whether this current param is a parent in bambi.

    Returns
    -------
    Param
        A Param object with info form param and model_config injected.
    """
    if isinstance(param, dict):
        name = param["name"]
        if "bounds" not in param:
            bounds = (
                model_config["bounds"].get(name, None)
                if "bounds" in model_config
                else None
            )
        else:
            bounds = param["bounds"]
        if "prior" not in param or param["prior"] is None:
            if (
                "default_priors" in model_config
                and name in model_config["default_priors"]
                and "formula" not in model_config
            ):
                prior = model_config["default_priors"][name]
            else:
                prior = None
        else:
            prior = param["prior"]
        return Param(
            name=name,
            prior=prior,
            formula=param.get("formula"),
            link=param.get("link"),
            bounds=bounds,
            is_parent=is_parent,
        )

    bounds = (
        model_config["bounds"].get(param, None) if "bounds" in model_config else None
    )
    prior = (
        model_config["default_priors"].get(param, None)
        if "default_priors" in model_config
        else None
    )
    return Param(
        name=param,
        prior=prior,
        bounds=bounds,
        is_parent=is_parent,
    )



"""Simulates data with basic ssm_simulators."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from ssms.basic_simulators import simulator
from ssms.config import model_config

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


def simulate_data(
    model: str,
    theta: ArrayLike,
    size: int,
    random_state: int | None = None,
    output_df: bool = True,
    **kwargs,
) -> np.ndarray | pd.DataFrame:
    """Sample simulated data from specified distributions.

    Parameters
    ----------
    model
        A model name that must be supported in `ssm_simulators`. For a detailed list of
        supported models, please see all fields in the `model_config` dict
        [here](https://github.com/AlexanderFengler/ssm-simulators/blob
        /e09eb2528d885c7b3340516597849fff4d9a5bf8/ssms/config/config.py#L6)
    theta
        An ArrayLike of floats that represent the true values of the parameters of the
        specified `model`. Please see [here](https://github.com/AlexanderFengler/
        ssm-simulators/blob/e09eb2528d885c7b3340516597849fff4d9a5bf8/ssms/config/config.py#L6)
        for what the parameters are. True values must be supplied in the same order
        as the parameters. You can also supply a 2D ArrayLike to simulate data for
        different trials with different true values.
    size
        The size of the data to be simulated. If `theta` is a 2D ArrayLike, this
        parameter indicates the size of data to be simulated for each trial.
    random_state : optional
        A random seed for reproducibility.
    output_df : optional
        If True, outputs a DataFrame with column names "rt", "response". Otherwise a
        2-column numpy array, by default True.
    kwargs : optional
        Other arguments passed to ssms.basic_simulators.simulator.

    Returns
    -------
    np.ndarray | pd.DataFrame
        An array or DataFrame with simulated data.
    """
    if model not in model_config:
        raise ValueError(f"model must be one of {list(model_config.keys())}.")

    sims = simulator(
        theta,
        model=model,
        n_samples=size,
        random_state=random_state,
        **kwargs,
    )

    rts = np.squeeze(sims["rts"]).reshape(-1)
    responses = np.squeeze(sims["choices"]).reshape(-1)

    sims_array = np.column_stack([rts, responses])

    if output_df:
        return pd.DataFrame(sims_array, columns=["rt", "response"])

    return sims_array



"""An ONNX to pytensor compiler.

Tips for extending this file to add more Ops:
If there is no equivalent in pytensor, then we need to write a function. The
`pytensor_gemm` function is an example of such a function. The number of argument is
the same as the number of inputs specified in the Opset document above. If there are
any optional inputs, also add these inputs as parameters with default values.
"""

import pytensor.tensor as pt

from .onnx2xla import _asarray, attribute_handlers


def pytensor_gemm(
    a, b, c=0.0, alpha=1.0, beta=1.0, transA=0, transB=0
):  # pylint: disable=C0103
    """Perform the GEMM op.

    Numpy-backed implementatio, of ONNX General Matrix Multiply (GeMM) op.
    """
    a = pt.transpose(a) if transA else a
    b = pt.transpose(b) if transB else b

    return [alpha * pt.dot(a, b) + beta * c]


pt_onnx_ops = {
    "Add": pt.add,
    "Constant": lambda value: [value],
    "MatMul": lambda x, y: [pt.dot(x, y)],
    "Relu": lambda x: [pt.math.max(x, 0)],
    "Reshape": lambda x, shape: [pt.reshape(x, shape)],
    "Tanh": lambda x: [pt.tanh(x)],
    "Gemm": pytensor_gemm,
}


def pt_interpret_onnx(graph, *args):
    """Transform model in onnx to pytensor.

    Parameters
    ----------
    graph
        The computation graph.
    args
        Inputs to the graph.

    Returns
    -------
        The result of the computation.
    """
    vals = dict(
        {n.name: a for n, a in zip(graph.input, args)},
        **{n.name: _asarray(n) for n in graph.initializer}
    )
    for node in graph.node:
        args = (vals[name] for name in node.input)
        attrs = {a.name: attribute_handlers[a.type](a) for a in node.attribute}
        outputs = pt_onnx_ops[node.op_type](*args, **attrs)
        for name, output in zip(node.output, outputs):
            vals[name] = output
    return [vals[n.name] for n in graph.output]
